# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Qwix for ODML."""

import dataclasses
import functools
from typing import Any, Callable, Sequence, Type

import flax
from flax import linen as nn
from flax import nnx
import jax
from jax import numpy as jnp
import numpy as np
from qwix._src import aux_data
from qwix._src import averaging
from qwix._src import flax_util
from qwix._src import qconfig
from qwix._src.core import qarray
from qwix._src.providers import odml_ops


class OdmlQatProvider(qconfig.QuantizationProvider):
  """QAT provider for ODML.

  Compared with the regular QAT provider, this provider
    * Quantizes all ops more than just conv, einsum, and dot_general.
    * Quantizes output activations via a delayed fake_quant.
    * Supports limited per-channel quantization for weights.
    * Doesn't support subchannel quantization.
  """

  def __init__(
      self,
      rules: Sequence[qconfig.QuantizationRule],
      *,
      disable_per_channel_weights: bool = False,
      fixed_range_for_inputs: tuple[float, float] | None = None,
      fixed_range_for_outputs: tuple[float, float] | None = None,
      strict: bool = True,
  ):
    """Constructor.

    Args:
      rules: The quantization rules.
      disable_per_channel_weights: Whether to disable per-channel quantization
        for weights.
      fixed_range_for_inputs: Use a fixed range when quantizing the model
        inputs, e.g. (0, 1).
      fixed_range_for_outputs: Use a fixed range when quantizing the model
        outputs, e.g. (0, 1).
      strict: Whether to raise an error if an unknown op is discovered.
    """
    super().__init__(rules)
    self._fixed_range_for_inputs = fixed_range_for_inputs
    self._fixed_range_for_outputs = fixed_range_for_outputs
    self._strict = strict
    self._ops = odml_ops.get_all_ops()

    for name in [
        'jax.lax.conv_general_dilated',
        'jax.lax.dot_general',
        'jax.numpy.einsum',
        'jax.numpy.dot',
    ]:
      self._ops[name] = functools.partial(
          self._ops[name],
          disable_per_channel_weights=disable_per_channel_weights,
          check_activation=strict,
      )

  def _init_rule(
      self, rule: qconfig.QuantizationRule
  ) -> qconfig.QuantizationRule:
    """Set ODML specific default values."""
    if rule.act_qtype is not None and rule.act_static_scale is None:
      rule = dataclasses.replace(rule, act_static_scale=True)
    if rule.act_calibration_method is None:
      rule = dataclasses.replace(rule, act_calibration_method='minmax')
    return super()._init_rule(rule)

  def nn_param(
      self,
      module: nn.Module,
      name: str,
      init_fn: Callable[..., Any],
      *init_args,
      unbox: bool = True,
      **init_kwargs,
  ) -> jax.Array | nn.meta.AxisMetadata[jax.Array]:
    """Intercepts nn.Module.param to associate weight_name aux_data."""
    ret = nn.Module.param(
        module, name, init_fn, *init_args, unbox=unbox, **init_kwargs
    )
    # Clear the previous aux_data such as fq_array.
    aux_data.clear(ret if unbox else ret.unbox())
    # weight_name is used to distinguish weights from activations.
    aux_data.set(ret if unbox else ret.unbox(), 'weight_name', name)
    return ret

  def get_intercept_map(self):
    """Used for interception."""
    intercept_map = super().get_intercept_map()
    intercept_map['flax.linen.Module.param'] = self.nn_param
    # Add all the ops to the intercept map.
    for name, op in self._ops.items():
      op: Type[odml_ops.QuantizedOp]
      intercept_map[name] = op(
          op_full_name=name,
          get_rule_and_op_id_fn=self._get_current_rule_and_op_id,
          fake_quant_fn=self._fake_quant,
      )
    return intercept_map

  def process_model_inputs(
      self, model: Any, model_args: Any, model_kwargs: Any
  ) -> tuple[Any, Any, Any]:
    """Quantize the input of the model."""
    # Set weight_name for nnx models. Linen models are handled in nn_param.
    if isinstance(model, nnx.Module):
      for path, node in nnx.iter_graph(model):
        if isinstance(node, nnx.Module):
          aux_data.clear(node)  # clear the op_count.
        elif isinstance(node, nnx.Param):
          # Clear the previous aux_data such as fq_array.
          aux_data.clear(node.value)
          # weight_name is used to distinguish weights from activations.
          aux_data.set(node.value, 'weight_name', path[-1])

    # Quantize the model inputs if needed.
    op = odml_ops.ModelInput(
        fixed_range_for_output=self._fixed_range_for_inputs,
        get_rule_and_op_id_fn=self._get_current_rule_and_op_id,
        fake_quant_fn=self._fake_quant,
    )
    return model, *jax.tree.map(op, (model_args, model_kwargs))

  def process_model_output(self, method_name: str, model_output: Any) -> Any:
    """Quantize the output of the model."""
    if method_name == '__call__':
      method_name = 'final'  # backwards compatibility.
    # Quantize the model output if needed.
    op = odml_ops.FinalOutput(
        op_full_name=method_name + '_output',
        fixed_range_for_output=self._fixed_range_for_outputs,
        get_rule_and_op_id_fn=self._get_current_rule_and_op_id,
        fake_quant_fn=self._fake_quant,
        check_activation=self._strict,
    )
    return jax.tree.map(op, model_output)

  def _fake_quant(
      self,
      array: jax.Array,
      how: qarray.HowToQuantize,
      quant_stat_name: str | None = None,
  ) -> jax.Array:
    """Apply fake quantization to array.

    This function can be used on both activations and weights. Gradient will be
    passed through.

    Args:
      array: The array to quantize.
      how: How to quantize the array.
      quant_stat_name: The name for the quantization statistics. If set, the
        quantization statistics will be collected and the scale will be computed
        from the statistics.

    Returns:
      The fake quantized array.
    """
    # Check and apply the fixed-range calibration asscociated with the array.
    fixed_range = aux_data.get(array, 'fixed_range', None)
    if fixed_range is not None:
      calibration_method = f'fixed,{fixed_range[0]},{fixed_range[1]}'
      how = dataclasses.replace(how, calibration_method=calibration_method)

    calibration = qarray.calibrate(array, how)
    if quant_stat_name is not None:
      is_fixed_range = how.calibration_method.startswith('fixed')
      calibration = self._collect_quant_stat(
          quant_stat_name, calibration, is_fixed_range
      )
    scale, zero_point = qarray.compute_scale_zero_point(calibration, how.qtype)
    q_array = qarray.quantize_with_scale_zero_point(
        array, how.qtype, scale, zero_point
    )
    dq_array = qarray.dequantize(q_array)
    # Use a straight through estimator as the gradient of the dq_array.
    ste_array = qarray.clip_to_calibration(array, calibration, how.tiled_axes)
    return ste_array + jax.lax.stop_gradient(dq_array - ste_array)

  def _collect_quant_stat(
      self,
      name: str,
      calibration: averaging.Calibration,
      calibration_is_fixed_range: bool,
  ) -> averaging.Calibration:
    """Collects the quantization statistics."""
    # For SRQ, only per-tensor scale is supported, so we don't need to check the
    # act_batch_axes at all.
    calibration = jax.tree.map(lambda x: x.mean(keepdims=True), calibration)

    aggregator = averaging.SimpleMovingAverage()
    quant_stat = flax_util.get_or_create_variable(
        'quant_stats', name, lambda: aggregator.init(calibration)
    )

    if flax_util.should_update_quant_stats():
      if calibration_is_fixed_range:
        # For fixed-range calibration, start from an empty quant_stat to avoid
        # floating-point accumulation error. Alternatively, we could skip
        # storing the quant_stat for fixed-range calibration.
        quant_stat.value = aggregator.init(calibration)
      quant_stat.value = aggregator.update(quant_stat.value, calibration)

    return aggregator.get_calibration(quant_stat.value, calibration)


class OdmlConversionProvider(OdmlQatProvider):
  """Quantization provider for ODML conversion.

  This mode is similar to OdmlQatProvider, but all fake_quant ops are annotated
  by composites and the scales are computed statically in numpy.

  Supported modes:
    * Weight-only quantization.
    * Static-range quantization.

  Usage:
    # The params can be from QAT or the FP model.
    params = ...

    # If using static-range quantization, quant_stats are needed and can be
    # obtained by either 1) QAT or 2) calibrating.
    quant_stats = ...

    # Apply OdmlConversionProvider to the model.
    conversion_model = qwix.quantize_model(
        fp_model, qwix.OdmlConversionProvider(rules, params, quant_stats)
    )
    # Convert and get the ODML model, which is an ai_edge_jax.model.TfLiteModel.
    odml_model = ai_edge_jax.convert(
        conversion_model.apply, {'params': params}, (inputs,)
    )
    # The odml_model can be exported or directly run.
    odml_model.export('/tmp/odml_model.tflite')
    odml_model(inputs)
  """

  def __init__(
      self,
      rules: Sequence[qconfig.QuantizationRule],
      params,
      quant_stats,
      **kwargs,
  ):
    super().__init__(rules, **kwargs)
    # Store params and quant_stats statically so they won't become tracers.
    self._flatten_params = flax.traverse_util.flatten_dict(params)
    self._quant_stats = quant_stats

  def get_intercept_map(self):
    intercept_map = super().get_intercept_map()
    # Override dot_general to flatten N-D weights to 2-D.
    intercept_map['jax.lax.dot_general'] = functools.partial(
        self._flatten_dot_general,
        _dot_general=intercept_map['jax.lax.dot_general'],
        _reshape=intercept_map['jax.lax.reshape'],
    )
    return intercept_map

  def _flatten_dot_general(self, *args, _dot_general, _reshape, **kwargs):
    """Flatten N-D weights to 2-D to support channelwise quantization."""
    # This special handling is needed because tflite doesn't support multiple
    # quantization_dimensions.
    if (
        aux_data.get(args[1], 'weight_name', None) is not None
        and args[1].ndim > 2
        and tuple(args[2][0][1]) == (0,)
    ):
      args = list(args)
      dout = args[1].shape[1:]
      args[1] = _reshape(args[1], (args[1].shape[0], np.prod(dout)))
      out = _dot_general(*args, **kwargs)
      return _reshape(out, out.shape[:-1] + dout)
    return _dot_general(*args, **kwargs)

  def _fake_quant(
      self,
      array: jax.Array,
      how: qarray.HowToQuantize,
      quant_stat_name: str | None = None,
  ) -> jax.Array:
    assert not how.tiled_axes, 'Tiled axes are not supported in ODML.'

    # Make the scale and zero point statically computed.
    with jax.ensure_compile_time_eval():
      # Check if the array is a weight or an activation.
      weight_name = aux_data.get(array, 'weight_name', None)
      if weight_name is not None:  # Weights.
        assert quant_stat_name is None
        mdl_path = flax_util.get_current_module_path()
        weight = self._flatten_params[mdl_path + (weight_name,)]
        if weight.shape != array.shape:  # when _flatten_dot_general is used.
          weight = weight.reshape(array.shape)
        calibration = qarray.calibrate(weight, how)
        scale, zp = qarray.compute_scale_zero_point(calibration, how.qtype)
      elif quant_stat_name is not None:  # Static-range activations.
        scale, zp = self._compute_static_scale_zero_point(how, quant_stat_name)
      else:  # Dynamic-range activations.
        scale, zp = None, None

      attributes = self._get_attributes(
          scale=scale, zp=zp, dtype=how.qtype, is_weight=weight_name is not None
      )

    @functools.partial(jax.lax.composite, name='quant.fake_quant')
    def _fake_quant_op(x, **attributes):
      del attributes  # attributes are only for the composite op.
      return qarray.dequantize(
          qarray.quantize(x, how)
          if scale is None
          else qarray.quantize_with_scale_zero_point(x, how.qtype, scale, zp)
      )

    return _fake_quant_op(array, **attributes)

  def _compute_static_scale_zero_point(
      self, how: qarray.HowToQuantize, quant_stat_name: str
  ) -> tuple[jax.Array, jax.Array | None]:
    """Statically compute the scale and zero point for weights or activations."""
    # Look up the quant_stat for the activation.
    obj = self._quant_stats
    for key in flax_util.get_current_module_path():
      obj = obj[key]
    quant_stat = obj[quant_stat_name]

    if 'count' not in quant_stat or quant_stat['count'] == 0:
      raise ValueError(f'quant_stats is not initialized for {quant_stat_name}')
    calibration = averaging.SimpleMovingAverage().get_calibration(quant_stat)
    return qarray.compute_scale_zero_point(calibration, how.qtype)

  def _get_attributes(
      self,
      *,
      scale: jax.Array | None,
      zp: jax.Array | None,
      dtype: jax.typing.DTypeLike,
      is_weight: bool,
  ) -> dict[str, Any]:
    """Return the attributes for the fake_quant composite."""
    # For dynamic-range quantization, the scale is an empty array.
    if scale is None:
      scale = np.array([], np.float32)
    if jnp.isnan(scale).any() or jnp.isinf(scale).any() or (scale == 0).any():
      raise ValueError(f'Invalid scale: {scale}')
    # Flatten the scale because ODML wants a 1D array.
    quantization_dim = None
    for dim, length in enumerate(scale.shape):
      if length > 1:
        if quantization_dim is None:
          quantization_dim = dim
        else:
          raise ValueError(f'Cannot flatten scale with shape {scale.shape}.')
    match jnp.dtype(dtype):
      case jnp.int8:
        dtype = 'i8'
      case _:
        raise ValueError(f'Unsupported dtype {dtype} for ODML conversion.')
    attributes = {
        'scale': np.asarray(scale, np.float32).flatten(),
        'dtype': dtype,
        # narrow_range is an ODML-specific optimization that reduces the range
        # of int8 quantization from [-128, 127] to [-127, 127], such that the
        # int8 x int8 product can be represented in int16. LiteRT quantization
        # spec requires narrow_range to be True for weights.
        #
        # Since Qwix uses [-127.5, 127.5] in symmetric quantization, setting
        # it to True will only affect exact -127.5 and should have minimal
        # impact on the quantization result.
        'narrow_range': is_weight,
    }
    if zp is not None:
      # zero_point has to be int64 for ODML.
      attributes['zero_point'] = np.asarray(zp, np.int64).flatten()
    if quantization_dim is not None:
      attributes['quantization_dimension'] = quantization_dim
    return attributes



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flax integration for QWIX."""

from collections.abc import Collection
import functools
import inspect
from typing import Any, TypeVar

from flax import linen as nn
from flax import nnx
from qwix._src import interception
from qwix._src import qconfig


ModelType = TypeVar("ModelType")


def quantize_model(
    model: ModelType,
    provider: qconfig.QuantizationProvider,
    *model_inputs: Any,
    methods: Collection[str] = ("__call__",),
    **model_inputs_kwargs: Any,
) -> ModelType:
  """Quantize a flax model.

  Args:
    model: The model to quantize, which can be previously quantized.
    provider: The quantization provider.
    *model_inputs: The inputs to the model, which should be provided if and only
      if the model is an NNX model.
    methods: The methods to quantize.
    **model_inputs_kwargs: The keyword arguments to the model, which should be
      provided if and only if the model is an NNX model.

  Returns:
    A quantized model instance.
  """
  if isinstance(model, nn.Module):
    if model_inputs or model_inputs_kwargs:
      raise ValueError("Model inputs must not be provided for linen models.")
    return quantize_linen_model(model, provider, methods)
  elif isinstance(model, nnx.Module):
    if not model_inputs and not model_inputs_kwargs:
      raise ValueError("Model inputs must be provided for nnx models.")
    if len(methods) != 1:
      raise ValueError("Only one method is supported for nnx models.")
    return quantize_nnx_model(
        model,
        provider,
        *model_inputs,
        call_method=next(iter(methods)),
        **model_inputs_kwargs,
    )
  else:
    raise ValueError(f"Unsupported model type: {type(model)}")


def quantize_linen_model(
    model: nn.Module,
    provider: qconfig.QuantizationProvider,
    methods: Collection[str],
) -> nn.Module:
  """Quantize a linen model."""

  def _is_in_nn_module() -> bool:
    nn_module: nn.Module = nn.module._context.module_stack[-1]  # pylint: disable=protected-access
    return nn_module and nn_module.scope is not None

  # Make a copy of the model to avoid modifying the original model.
  model = model.copy()

  # We need to modify the model class's methods because nn.Module.apply will
  # invoke a copy of the model rather than the original one, so setting the
  # attributes of the model doesn't always work. For example, in a VAE model,
  # we intercept the encode and decode methods, but invoke the __call__ method
  # during training. The interception on encode and decode methods will be
  # unset when the model is copied.
  model_class = model.__class__
  # If the model class is already quantized, get the unquantized type.
  if hasattr(model_class, "_unquantized_type"):
    model_class = model_class._unquantized_type  # pylint: disable=protected-access

  new_fields = {"_unquantized_type": model_class}
  for method_name in methods:
    method = getattr(model_class, method_name)

    # Step 1: unwrap if needed.
    wrapped = False
    # This is a while loop because the method may be wrapped e.g. by nn.jit as
    # jit(wrap_method_once(method)). We discard the jit decorator for now.
    while hasattr(method, "method_handler_wrapped"):
      # Apply interception to the unwrapped method so that input/output
      # transformations are called inside a proper linen scope.
      method = method.__wrapped__
      wrapped = True

    # Step 2: intercept the method.
    method = interception.wrap_func_intercepted(
        method,
        provider.get_intercept_map,
        input_transform=functools.partial(_input_transform, provider),
        output_transform=functools.partial(
            provider.process_model_output, method_name
        ),
        should_intercept=_is_in_nn_module,
    )

    # Step 3: wrap if needed.
    if wrapped:
      method = nn.module.wrap_method_once(method)

    new_fields[method_name] = method

  # Create a new class for the model.
  model.__class__ = type(model_class.__name__, (model_class,), new_fields)
  return model


def quantize_nnx_model(
    model: nnx.Module,
    provider: qconfig.QuantizationProvider,
    *model_inputs: Any,
    call_method: str = "__call__",
    **model_inputs_kwargs: Any,
) -> nnx.Module:
  """Quantize an NNX model.

  To fully quantize an NNX model, Qwix needs to run the model at least once.
  Thus model inputs are required. The actual values do not matter.

  If the model already contains the correct original weights, this function will
  quantize them correctly.

  Args:
    model: An NNX model instance, which can be previously quantized.
    provider: The quantization provider.
    *model_inputs: The inputs to the model.
    call_method: The name of the method to call, default to __call__.
    **model_inputs_kwargs: The keyword arguments to the model.

  Returns:
    A quantized model instance.
  """
  # Make a copy of the model to avoid modifying the original model.
  model = nnx.clone(model)

  model_class = model.__class__
  # If the model class is already quantized, get the unquantized type.
  if hasattr(model_class, "_unquantized_type"):
    model_class = model_class._unquantized_type  # pylint: disable=protected-access

  # Only intercepts the `call_method` of the nnx module for now.
  #
  # The class method type(model).__call__ will be invoked when calling the
  # instance of the class so we create a new class with the intercepted __call__
  # and the update the class of the model.
  new_fields = {
      "_unquantized_type": model_class,
      call_method: interception.wrap_func_intercepted(
          getattr(model_class, call_method),
          provider.get_intercept_map,
          input_transform=functools.partial(_input_transform, provider),
          output_transform=functools.partial(
              _output_transform_nnx, provider, call_method
          ),
      ),
  }
  model.__class__ = type(model_class.__name__, (model_class,), new_fields)

  # Unlike linen module, nnx module does not have scope or path attribute, we
  # need to iterate over all modules and set the path for them.
  for path, module in model.iter_modules():
    module.qwix_path = path
    # Disable quant_stats update for the first call.
    module.disable_quant_stats_update = True

  # Because nnx modules are stateful, we need to call them once to initialize
  # them (convert weights, create quant_stats) unless users explicitly opt out.
  if not model_inputs_kwargs.get("skip_nnx_init"):
    getattr(model, call_method)(*model_inputs, **model_inputs_kwargs)

  # Enable quant_stats update for subsequent calls.
  for _, module in model.iter_modules():
    module.disable_quant_stats_update = False

  return model


def _input_transform(provider: qconfig.QuantizationProvider, args, kwargs):
  model, *args = args
  model, args, kwargs = provider.process_model_inputs(model, args, kwargs)
  return (model, *args), kwargs


def _output_transform_nnx(
    provider: qconfig.QuantizationProvider, method_name: str, output: Any
):
  # Create a frame with a variable named `self` pointing to the model so that
  # flax_util.get_current_module() can work inside the output transform.
  # We cannot use the model in quantize_nnx_model because users may choose to
  # clone the model.
  args = inspect.currentframe().f_back.f_locals["args"]  # pytype: disable=attribute-error
  self = args[0]  # pylint: disable=unused-variable
  return provider.process_model_output(method_name, output)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized jax.lax.conv_general_dilated."""

from collections.abc import Sequence
from typing import Any
import jax
from jax import numpy as jnp
from qwix._src.core import numerics
from qwix._src.core import qarray


def get_how_to_quantize(
    *,
    dimension_numbers: jax.lax.ConvDimensionNumbers,
    for_lhs: bool,
    **kwargs: Any,
) -> qarray.HowToQuantize:
  """Gets how to quantize from conv's dimension_numbers.

  Use channelwise for batch dimension and out feature dimension.

  Args:
    dimension_numbers: The conv's dimension_numbers.
    for_lhs: Whether to quantize lhs or rhs.
    **kwargs: Additional keyword arguments to HowToQuantize.

  Returns:
    How to quantize lhs or rhs.
  """
  if for_lhs:
    channelwise_axes = [dimension_numbers.lhs_spec[0]]
  else:
    channelwise_axes = [dimension_numbers.rhs_spec[0]]
  return qarray.HowToQuantize(
      channelwise_axes=channelwise_axes,
      tiled_axes={},
      **kwargs,
  )


def get_transpose(
    dimension_numbers: jax.lax.ConvDimensionNumbers, for_lhs: bool
) -> list[int | None]:
  """Returns the transpose list for the given dimension_numbers."""
  transpose = [None] * len(dimension_numbers.out_spec)
  if for_lhs:
    # Only batch dimension can be channelwise thus transposed.
    transpose[dimension_numbers.out_spec[0]] = dimension_numbers.lhs_spec[0]
  else:
    # Only out feature dimension can be channelwise thus transposed.
    transpose[dimension_numbers.out_spec[1]] = dimension_numbers.rhs_spec[0]
  return transpose


def _slow_conv_general_dilated(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    window_strides: Sequence[int],
    padding: str | Sequence[tuple[int, int]],
    lhs_dilation: Sequence[int] | None = None,
    rhs_dilation: Sequence[int] | None = None,
    dimension_numbers: jax.lax.ConvGeneralDilatedDimensionNumbers = None,
    feature_group_count: int = 1,
    batch_group_count: int = 1,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    out_sharding=None,
) -> jax.Array:
  """Dequantizes first then computes in floating-point types."""
  if isinstance(lhs, qarray.QArray):
    lhs = qarray.dequantize(lhs)
  if isinstance(rhs, qarray.QArray):
    rhs = qarray.dequantize(rhs)
  return jax.lax.conv_general_dilated(
      lhs,
      rhs,
      window_strides,
      padding,
      lhs_dilation,
      rhs_dilation,
      dimension_numbers,
      feature_group_count,
      batch_group_count,
      precision,
      preferred_element_type,
      out_sharding,
  )


def _fast_conv_general_dilated(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    window_strides: Sequence[int],
    padding: str | Sequence[tuple[int, int]],
    lhs_dilation: Sequence[int] | None = None,
    rhs_dilation: Sequence[int] | None = None,
    dimension_numbers: jax.lax.ConvGeneralDilatedDimensionNumbers = None,
    feature_group_count: int = 1,
    batch_group_count: int = 1,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    out_sharding=None,
) -> jax.Array:
  """Quantized jax.lax.conv_general_dilated. Both sides must be QArrays."""
  dimension_numbers = jax.lax.conv_dimension_numbers(
      lhs.shape, rhs.shape, dimension_numbers
  )
  preferred_element_type, result_type = qarray.get_accumulator_and_result_type(
      lhs, rhs, preferred_element_type=preferred_element_type
  )

  if isinstance(lhs, qarray.QArray):
    lhs_value = lhs.qvalue
    lhs_scale = lhs.scale
    lhs_zero_point = lhs.zero_point
    if qarray.get_tiled_axes(lhs):
      raise ValueError('subchannel is not supported for conv_general_dilated.')
  else:
    lhs_value = lhs
    lhs_scale = None
    lhs_zero_point = None

  if isinstance(rhs, qarray.QArray):
    rhs_value = rhs.qvalue
    rhs_scale = rhs.scale
    rhs_zero_point = rhs.zero_point
    if qarray.get_tiled_axes(rhs):
      raise ValueError('subchannel is not supported for conv_general_dilated.')
  else:
    rhs_value = rhs
    rhs_scale = None
    rhs_zero_point = None

  if rhs_zero_point is not None:
    raise ValueError('Asymmetric quantization for rhs is not supported.')

  res = jax.lax.conv_general_dilated(
      lhs_value,
      rhs_value,
      window_strides=window_strides,
      padding=padding,
      lhs_dilation=lhs_dilation,
      rhs_dilation=rhs_dilation,
      dimension_numbers=dimension_numbers,
      feature_group_count=feature_group_count,
      batch_group_count=batch_group_count,
      preferred_element_type=preferred_element_type,
      out_sharding=out_sharding,
  )
  if lhs_zero_point is not None:
    # TODO(zhuyunx): This value can be constant folded in SRQ scenarios.
    res -= jax.lax.conv_general_dilated(
        jnp.broadcast_to(lhs_zero_point, lhs_value.shape),
        rhs_value,
        window_strides=window_strides,
        padding=padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
        preferred_element_type=preferred_element_type,
    )

  if lhs_scale is not None:
    transpose = get_transpose(dimension_numbers, for_lhs=True)
    lhs_scale = qarray.transpose_array(lhs_scale, transpose)
    res = qarray.call_with_generic_broadcast(jnp.multiply, res, lhs_scale)

  if rhs_scale is not None:
    transpose = get_transpose(dimension_numbers, for_lhs=False)
    rhs_scale = qarray.transpose_array(rhs_scale, transpose)
    res = qarray.call_with_generic_broadcast(jnp.multiply, res, rhs_scale)

  return res.astype(result_type)


def conv_general_dilated(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    window_strides: Sequence[int],
    padding: str | Sequence[tuple[int, int]],
    lhs_dilation: Sequence[int] | None = None,
    rhs_dilation: Sequence[int] | None = None,
    dimension_numbers: jax.lax.ConvGeneralDilatedDimensionNumbers = None,
    feature_group_count: int = 1,
    batch_group_count: int = 1,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    out_sharding=None,
) -> jax.Array:
  """Dispatches to fast or slow conv_general_dilated depending on the inputs."""
  use_fast_path = True

  for x in (lhs, rhs):
    if isinstance(x, jax.Array) and numerics.should_quantize(x.dtype):
      use_fast_path = False
      break

  if use_fast_path:
    return _fast_conv_general_dilated(
        lhs,
        rhs,
        window_strides,
        padding,
        lhs_dilation,
        rhs_dilation,
        dimension_numbers,
        feature_group_count,
        batch_group_count,
        preferred_element_type,
        out_sharding,
    )
  else:
    return _slow_conv_general_dilated(
        lhs,
        rhs,
        window_strides,
        padding,
        lhs_dilation,
        rhs_dilation,
        dimension_numbers,
        feature_group_count,
        batch_group_count,
        precision,
        preferred_element_type,
        out_sharding,
    )



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Qwix: A Jax quantization library."""

# pylint: disable=g-multiple-import, g-importing-member

from qwix._src.core.conv_general import conv_general_dilated
from qwix._src.core.dot import dot
from qwix._src.core.dot_general import dot_general
from qwix._src.core.einsum import einsum
from qwix._src.core.qarray import QArray, dequantize, quantize_api as quantize
from qwix._src.core.ragged_dot import ragged_dot, ragged_dot_general
from qwix._src.flax_util import QuantStat
from qwix._src.model import quantize_model
from qwix._src.providers.lora import LoraProvider, LoraRule, apply_lora_to_model
from qwix._src.providers.odml import OdmlConversionProvider, OdmlQatProvider
from qwix._src.providers.ptq import PtqProvider, quantize_params
from qwix._src.providers.qt import QtProvider, QtRule
from qwix._src.qconfig import QuantizationProvider, QuantizationRule



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized einsum with subchannel support."""

import dataclasses
from typing import Any, Collection

import jax
from jax import numpy as jnp
from qwix._src.core import dot_general
from qwix._src.core import qarray


@dataclasses.dataclass(slots=True)
class EinsumInfo:
  lhs: str
  rhs: str
  out: str
  contractions: Collection[str]


def get_einsum_info(einsum_str: str, ndims: tuple[int, int]) -> EinsumInfo:
  """Gets einsum info from an einsum string."""
  einsum_str = einsum_str.replace(' ', '')
  inputs, out = einsum_str.split('->')
  lhs, rhs = inputs.split(',')
  if '...' in lhs or '...' in rhs:
    ndim = ndims[0] - len(lhs) + 3 if '...' in lhs else ndims[1] - len(rhs) + 3
    assert ndim <= 10, f'{ndim=} {einsum_str=}'
    digits = ''.join(map(str, range(ndim)))
    assert not set(digits) & set(einsum_str), f'{digits=} {einsum_str=}'
    lhs = lhs.replace('...', digits)
    rhs = rhs.replace('...', digits)
    out = out.replace('...', digits)
  return EinsumInfo(lhs, rhs, out, set(lhs) & set(rhs) - set(out))


def get_how_to_quantize(
    *,
    einsum_str: str,
    ndims: tuple[int, int],
    for_lhs: bool,
    tile_size: int | float | None,
    **kwargs: Any,
) -> qarray.HowToQuantize:
  """Get how to quantize from an einsum string.

  By default, use channelwise for all non-contraction axes, and subchannel for
  contraction axes if a tile_size is given.

  Args:
    einsum_str: The einsum string.
    ndims: The number of dimensions of the lhs and rhs array. This is needed
      when ellipsis is in subscripts and we need to determine the number of
      dimensions represented by ellipsis.
    for_lhs: Whether to quantize lhs or rhs.
    tile_size: The tile size for subchannel quantization.
    **kwargs: Additional keyword arguments to HowToQuantize.

  Returns:
    How to quantize the lhs or rhs.
  """
  info = get_einsum_info(einsum_str, ndims)
  subs = info.lhs if for_lhs else info.rhs
  channelwise_axes = []
  tiled_axes = {}
  for axis, name in enumerate(subs):
    if name not in info.contractions:
      channelwise_axes.append(axis)
    elif tile_size and not tiled_axes:  # Only tile the first contraction axis.
      tiled_axes[axis] = tile_size

  return qarray.HowToQuantize(
      channelwise_axes=channelwise_axes,
      tiled_axes=tiled_axes,
      **kwargs,
  )


def einsum(
    *args,
    _qwix_dot_general=dot_general.dot_general,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    **kwargs,
) -> jax.Array:
  """Quantized einsum that can take QArrays and returns floating-point jax.Array.

  Args:
    *args: Arguments to einsum.
    _qwix_dot_general: The dot_general function to use.
    preferred_element_type: The preferred element type for jax.lax.dot_general.
    **kwargs: Keyword arguments to einsum.

  Returns:
    The result of the einsum, a floating-point jax.Array.
  """
  # preferred_element_type has to be set for jnp.einsum so that it won't infer
  # the type from qvalue x qvalue.
  _, preferred_element_type = qarray.get_accumulator_and_result_type(
      *[a for a in args if isinstance(a, qarray.MaybeQArray)],
      preferred_element_type=preferred_element_type,
  )

  # We want to use jnp.einsum with quantized dot_general to avoid duplicating
  # the implementation. However, jnp.einsum will check the inputs to be
  # jax Arrays. To work around this, we send the qvalue to jnp.einsum and
  # restore the actual QArray before actually passing them to dot_general.
  args = list(args)
  qvalue_to_qarray = {}
  for i, arg in enumerate(args):
    if isinstance(arg, qarray.QArray):
      args[i] = arg.qvalue
      qvalue_to_qarray[id(arg.qvalue)] = arg

  def _dot_general(*args, **kwargs):
    args = [qvalue_to_qarray.pop(id(a), a) for a in args]
    return _qwix_dot_general(*args, **kwargs)

  # Disabling JIT is necessary so that args in _dot_general are not tracers.
  with jax.disable_jit():
    out = jnp.einsum(
        *args,
        _dot_general=_dot_general,
        preferred_element_type=preferred_element_type,
        **kwargs,
    )
  assert not qvalue_to_qarray, 'All qvalues should be consumed.'
  return out



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""ODML ops for QAT."""

import dataclasses
import functools
import sys
from typing import Any, Callable, Sequence

import jax
from jax import numpy as jnp
from qwix._src import aux_data
from qwix._src import qconfig
from qwix._src.core import conv_general
from qwix._src.core import dot_general
from qwix._src.core import einsum
from qwix._src.core import qarray


def get_all_ops():
  """Get all the functions to intercept and the corresponding ops."""

  # Quantization ops can generally be categorized into 3 categories:
  # https://ai.google.dev/edge/litert/models/quantization_spec
  #
  #   1. Ops that allow different scales for inputs and output. For example,
  #      add, mean, etc. These ops have FQ on all inputs and output. They are
  #      handled either by the generic QuantizedOp class (e.g. add), or by
  #      custom classes inheriting from QuantizedOp (e.g. dot_general).
  #   2. Ops that only allow the same scale for inputs and output. For example,
  #      reshape, avg_pool, split, etc. These ops have FQ only on the input or
  #      the output, and the converter is able to propagate the scale.
  #      1. FQ on the input: avg_pool, resize, etc. Their output values depend
  #         on the complete input values.
  #      2. FQ on the output: clip, pad, relu, etc. Their output values don't
  #         depend on the unused part of the input values, so it's more accurate
  #         to use the scale from the output.
  #      3. Some "transparent" ops such as reshape, transpose, etc. don't change
  #         the actual value of the array. They cannot be selected by the
  #         quantization rules. Instead, their behavior depends on whether the
  #         previous op is quantized.
  #   3. Ops that don't have a corresponding quantized op, e.g. sin, cos. If
  #      the previous op wants to quantize the output, the op will fake quantize
  #      its input. Otherwise, the op will insert no FQ.

  # l2_norm is not listed here because there's no standard implementation.
  # Usually people implement it as
  #
  #   def l2_norm(x, eps=1e-6):
  #     return x / jnp.maximum(jnp.linalg.norm(x, axis=-1, keepdims=True), eps)
  #
  # To support this op, register it manually in the provider (both QAT and
  # conversion) using the tanh handler, e.g.
  #
  #   l2_norm_full_name = l2_norm.__module__ + '.' + l2_norm.__name__
  #   provider._ops[l2_norm_full_name] = provider._ops['jax.numpy.tanh']

  partial = functools.partial
  quantize = lambda *a, **k: functools.partial(QuantizedOp, input_idx=a, **k)

  return {
      # go/keep-sorted start
      'flax.linen.BatchNorm.__call__': BatchNorm,
      'flax.linen.Dropout.__call__': partial(TransparentOp, input_idx=[1]),
      'flax.linen.GroupNorm.__call__': quantize(1, op_name='norm_op'),
      'flax.linen.LayerNorm.__call__': quantize(1, op_name='norm_op'),
      'flax.linen.avg_pool': OnlyInputOp,
      'flax.linen.max_pool': OnlyInputOp,
      'jax._src.numpy.indexing.rewriting_take': Take,  # a.__getitem__
      'jax.custom_jvp.__call__': CustomJvpCall,  # handles relu and relu6.
      'jax.image.resize': OnlyInputOp,
      'jax.lax.broadcast_in_dim': TransparentOp,
      'jax.lax.conv_general_dilated': DotEinsumConv,
      'jax.lax.dot_general': DotEinsumConv,
      'jax.lax.reshape': TransparentOp,
      'jax.lax.split': Split,
      'jax.lax.squeeze': TransparentOp,
      'jax.lax.stop_gradient': TransparentOp,
      'jax.lax.transpose': TransparentOp,
      'jax.lax.with_sharding_constraint': TransparentOp,
      'jax.nn.gelu': quantize(0),
      'jax.nn.leaky_relu': quantize(0),
      'jax.nn.silu': Silu,
      'jax.nn.softmax': Softmax,
      'jax.numpy.array': TransparentOp,
      'jax.numpy.astype': TransparentOp,
      'jax.numpy.clip': OnlyOutputOp,
      'jax.numpy.concatenate': Concatenate,
      'jax.numpy.cos': NoQuantOp,
      'jax.numpy.dot': DotEinsumConv,
      'jax.numpy.einsum': DotEinsumConv,
      'jax.numpy.equal': functools.partial(NoQuantOp, input_idx=[0, 1]),
      'jax.numpy.floor': OnlyOutputOp,
      'jax.numpy.mean': quantize(0),
      'jax.numpy.pad': OnlyOutputOp,
      'jax.numpy.repeat': quantize(0),  # not fully supported by the converter.
      'jax.numpy.sin': NoQuantOp,
      'jax.numpy.squeeze': TransparentOp,
      'jax.numpy.sum': quantize(0),
      'jax.numpy.take': Take,
      'jax.numpy.tanh': Tanh,
      'jax.numpy.true_divide': quantize(0, 1, op_name='truediv'),
      'jax.numpy.ufunc.__call__': UfuncCall,  # handles add, sub, mul, pow.
      # go/keep-sorted end
  }


NotAnActivationError = ValueError(
    "The array is expected to be an activation, but it's not. This is usually"
    " because your model has ops that Qwix doesn't support, or your model is"
    ' using those ops in a way Qwix failed to intercept. To disable this check,'
    ' pass `strict=False` to the provider.'
)


### Possible auxiliary data associated with an array

# Whether an array should be fake quantized by the next op and what rule to use.
#
# For the output of an op, it's not fake-quantized immediately because the next
# op may choose to delay the FQ, e.g. dot_general + add + relu can be fused and
# no FQ should be inserted in between.
_FQ_RULE = 'fq_rule'  # QuantizationRule

# Whether the (unquantized) array is already fake-quantized in another code path
# and what the fake-quantized array is. This avoids the same array being
# fake-quantized multiple times.
_FQ_ARRAY = 'fq_array'  # array

# Whether the previous op allows to fuse arithmetic ops or batch norm after it.
_ALLOW_FUSION = 'allow_fusion'  # bool

# Whether the array is an activation. An array can be either an activation,
# a weight, or a constant.
_IS_ACTIVATION = 'is_activation'  # bool

# Whether the array is a weight and what is its name. Weights don't need to have
# quantization statistics collected because they are statically quantized.
# The name is useful in the conversion provider to find the static weight.
_WEIGHT_NAME = 'weight_name'  # str

# Fixed range for logistic functions whose output ranges are known, e.g.
# softmax.
_FIXED_RANGE = 'fixed_range'  # tuple[float, float]


GetRuleAndOpIdFn = Callable[[str], tuple[qconfig.QuantizationRule, str]]
FakeQuantFn = Callable[[jax.Array, qarray.HowToQuantize, str | None], jax.Array]


class QuantizedOp:
  """A generic quantized op that allows different scales for inputs and output.

  Each op is initialized once and invoked multiple times. Thus the rule and the
  op id are given as a function.
  """

  # Which args are considered as the inputs of the op.
  input_idx: Sequence[int] = ()

  # Fixed range for the op output.
  fixed_range_for_output: tuple[float, float] | None = None

  def __init__(
      self,
      *,
      op_full_name: str,
      get_rule_and_op_id_fn: GetRuleAndOpIdFn,
      fake_quant_fn: FakeQuantFn,
      op_name: str | None = None,
      **kwargs,
  ):
    """Create a generic op.

    Arguments here are provided during runtime.

    Args:
      op_full_name: The full name of the op, e.g. jax.numpy.add.
      get_rule_and_op_id_fn: A function that returns the quantization rule and
        the op id for the op.
      fake_quant_fn: A function that performs fake quantization on an array.
      op_name: The name of the op, e.g. add. If not given, it will be inferred
        from op_full_name.
      **kwargs: Additional keyword arguments to override class attributes.
    """
    # Those are attributes that are given during runtime.
    self._op_full_name = op_full_name
    self._get_rule_and_op_id_fn = get_rule_and_op_id_fn
    self._fake_quant_fn = fake_quant_fn
    if op_name is None:
      op_name = op_full_name.split('.')[-1].replace('__', '')
    self._op_name = op_name

    # Use kwargs to override class attributes.
    for key, value in kwargs.items():
      if not hasattr(type(self), key):
        raise AttributeError(f'Unsupported keyword argument {key}.')
      setattr(self, key, value)

  def __call__(self, *args, **kwargs):
    """Quantize the op."""
    # If there's no activation in the args, skip the quantization because
    # they are all weights or constants that can be constant-folded.
    if not self._inputs_have_activations(args):
      return self._call_original_op(*args, **kwargs)

    rule, op_id = self._get_rule_and_op_id_fn(self._op_name)
    args = self._fake_quant_inputs(args, rule, op_id)
    out = self._call_original_op(*args, **kwargs)
    return self._fake_quant_output(out, rule)

  def _inputs_have_activations(self, args: Sequence[Any]) -> bool:
    """Check if there's any activation in the inputs."""
    if not self.input_idx:
      raise ValueError(f'input_idx is not set for op {self._op_name}.')
    for idx in self.input_idx:
      if isinstance(args[idx], jax.Array) and aux_data.get(
          args[idx], _IS_ACTIVATION, False
      ):
        return True
    return False

  def _call_original_op(self, *args, **kwargs) -> Any:
    """Get the original function from op_full_name."""
    name_parts = self._op_full_name.split('.')
    obj = sys.modules[name_parts[0]]
    for attr in name_parts[1:]:
      obj = getattr(obj, attr)
    return obj(*args, **kwargs)

  def _fake_quant_inputs(
      self,
      args: Sequence[Any],
      rule: qconfig.QuantizationRule | None,
      op_id: str,
  ) -> Sequence[Any]:
    """Fake quantize the inputs of the op."""
    args = list(args)
    if len(self.input_idx) == 1:
      idx = self.input_idx[0]
      args[idx] = self._maybe_fake_quant(args[idx], rule, op_id)
    elif len(self.input_idx) == 2:
      lhs, rhs = tuple(self.input_idx)  # pylint: disable=unbalanced-tuple-unpacking
      # Binary ops could have non-array args, e.g. x + 1.
      if isinstance(args[lhs], jax.Array):
        args[lhs] = self._maybe_fake_quant(args[lhs], rule, op_id + '_lhs')
      if isinstance(args[rhs], jax.Array):
        args[rhs] = self._maybe_fake_quant(args[rhs], rule, op_id + '_rhs')
    else:
      raise ValueError(
          f'Unsupported num of inputs {self.input_idx} for op {self._op_name}.'
      )
    return args

  def _maybe_fake_quant(
      self,
      array: jax.Array,
      rule: qconfig.QuantizationRule | None,
      quant_stat_name: str,
  ) -> jax.Array:
    """Fake quantize the array based on the given rule.

    This function assumes the array is an activation, unless it has weight_name
    aux_data, e.g., in jnp.take.

    Args:
      array: The array to quantize.
      rule: The quantization rule for the array. If None, the array will not be
        quantized.
      quant_stat_name: The name for the quantization statistics.

    Returns:
      The fake quantized array.
    """
    # Check if the array is already quantized in another code path.
    fq_array = aux_data.get(array, _FQ_ARRAY, None)
    if fq_array is not None:
      return array if fq_array == 'self' else fq_array

    # Only quantize float arrays.
    if array.dtype not in (jnp.float32, jnp.bfloat16):
      return array

    # Check if the array is a weight.
    if aux_data.get(array, _WEIGHT_NAME, None) is not None:
      if rule and rule.weight_qtype:
        how = qarray.HowToQuantize(
            qtype=rule.weight_qtype,
            channelwise_axes=(),
            tiled_axes={},
            # Use act_calibration_method because it is more like an activation,
            # i.e., asymmetric rather than symmetric.
            calibration_method=rule.act_calibration_method,
        )
        fq_array = self._fake_quant_fn(array, how, None)
        aux_data.set(array, _FQ_ARRAY, fq_array)
        return fq_array
      return array

    # Check if the array should be quantized as the output of the previous op.
    previous_rule = aux_data.get(array, _FQ_RULE, None)
    if previous_rule is not None:
      rule = previous_rule

    if rule is None or rule.act_qtype is None:
      return array
    if not rule.act_static_scale:
      # DRQ is only supported in DotEinsumConv and they should call
      # _fake_quant_fn directly.
      return array

    how = qarray.HowToQuantize(
        qtype=rule.act_qtype,
        tiled_axes={},
        # Use per-channel scales for batch axes, which will be reduced later
        # in _collect_quant_stat.
        channelwise_axes=rule.act_batch_axes,
        calibration_method=rule.act_calibration_method,
    )

    fq_array = self._fake_quant_fn(array, how, quant_stat_name)
    aux_data.set(array, _FQ_ARRAY, fq_array)
    return fq_array

  def _fake_quant_output(
      self, array: jax.Array, rule: qconfig.QuantizationRule | None
  ) -> jax.Array:
    """Fake quantize an output activation, which is delayed to the next op."""
    aux_data.set(array, _IS_ACTIVATION, True)
    if self.fixed_range_for_output is not None:
      aux_data.set(array, _FIXED_RANGE, self.fixed_range_for_output)
    # Output is only quantized in SRQ.
    if rule and rule.act_qtype and rule.act_static_scale:
      aux_data.set(array, _FQ_RULE, rule)
    return array


class OnlyInputOp(QuantizedOp):
  """An op that only fake quantizes the input."""

  input_idx = [0]  # default to a unary op.

  def __call__(self, *args, **kwargs):
    if not self._inputs_have_activations(args):
      return self._call_original_op(*args, **kwargs)
    rule, op_id = self._get_rule_and_op_id_fn(self._op_name)
    # Only fake quantize the input.
    args = self._fake_quant_inputs(args, rule, op_id)
    out = self._call_original_op(*args, **kwargs)
    if rule and rule.act_qtype:
      # Mark the output as already quantized.
      aux_data.set(out, _FQ_ARRAY, 'self')
    return self._fake_quant_output(out, rule)


class OnlyOutputOp(QuantizedOp):
  """An op that only fake quantizes the output."""

  input_idx = [0]  # default to a unary op.

  def __call__(self, *args, **kwargs):
    if not self._inputs_have_activations(args):
      return self._call_original_op(*args, **kwargs)
    rule, _ = self._get_rule_and_op_id_fn(self._op_name)
    if rule is None or rule.act_qtype is None:
      rule = aux_data.get(args[self.input_idx[0]], _FQ_RULE, None)
    # No quantization on the input.
    out = self._call_original_op(*args, **kwargs)
    return self._fake_quant_output(out, rule)


class TransparentOp(QuantizedOp):
  """A transparent op doesn't quantize anything. It only forwards aux data.

  This is mainly used for ops that doesn't change the actual value of the
  inputs, e.g. reshape, transpose, etc. This is similar to OnlyOutputOp,
  but it reuses the previous rule and forwards more aux data.
  """

  input_idx = [0]  # default to a unary op.
  forwarded_aux_data = (
      _IS_ACTIVATION,
      _WEIGHT_NAME,
      _FQ_RULE,
      _FIXED_RANGE,
      _ALLOW_FUSION,
  )

  def __call__(self, *args, **kwargs):
    if len(self.input_idx) > 1:
      raise ValueError(
          f'Unsupported num of inputs {self.input_idx} for op {self._op_name}.'
      )
    out = self._call_original_op(*args, **kwargs)

    def forward(out, arg):
      for key in self.forwarded_aux_data:
        value = aux_data.get(arg, key, None)
        if value is not None:
          aux_data.set(out, key, value)
      # Also forward the FQ_ARRAY if it's used to skip the quantization.
      fq_array = aux_data.get(arg, _FQ_ARRAY, None)
      if fq_array == 'self':
        aux_data.set(out, _FQ_ARRAY, fq_array)

    jax.tree.map(forward, out, args[self.input_idx[0]])
    return out


class NoQuantOp(QuantizedOp):
  """An fp op doesn't have a corresponding quantized op."""

  input_idx = [0]  # default to a unary op.

  def __call__(self, *args, **kwargs):
    if not self._inputs_have_activations(args):
      return self._call_original_op(*args, **kwargs)

    # Only fake quantize the input when the previous op wants.
    _, op_id = self._get_rule_and_op_id_fn(self._op_name)
    args = self._fake_quant_inputs(args, None, op_id)
    out = self._call_original_op(*args, **kwargs)

    # and don't quantize the output.
    return self._fake_quant_output(out, None)


class ModelInput(QuantizedOp):
  """A synthetic op for the model input."""

  def __init__(self, **kwargs):
    # The op_full_name doesn't matter here.
    super().__init__(op_full_name='model_input', **kwargs)

  def __call__(self, x: Any) -> Any:
    aux_data.clear(x)  # such as FQ_ARRAY
    if isinstance(x, jax.Array):
      self._fake_quant_output(x, None)
    return x


class FinalOutput(QuantizedOp):
  """A synthetic op for the model output."""

  # Whether to check if the output is an activation.
  check_activation: bool = True

  def __init__(self, **kwargs):
    super().__init__(**kwargs)

  def __call__(self, x: Any) -> Any:
    if self.check_activation and not aux_data.get(x, _IS_ACTIVATION, False):
      raise NotAnActivationError
    _, op_id = self._get_rule_and_op_id_fn(self._op_name)
    if self.fixed_range_for_output is not None:
      aux_data.set(x, _FIXED_RANGE, self.fixed_range_for_output)
    # Only FQ the output if the previous op wants.
    return self._maybe_fake_quant(x, None, op_id)


class BatchNorm(QuantizedOp):
  """BatchNorm op, which can be fused into previous op completely."""

  def __call__(self, norm, x: jax.Array, *args, **kwargs) -> jax.Array:
    if not aux_data.get(x, _IS_ACTIVATION, False):
      return norm(x, *args, **kwargs)
    if aux_data.get(x, _ALLOW_FUSION, False):
      rule = aux_data.get(x, _FQ_RULE, None)
      out = norm(x, *args, **kwargs)
      aux_data.set(out, _ALLOW_FUSION, True)
    else:
      rule, op_id = self._get_rule_and_op_id_fn('batch_norm_op')
      x = self._maybe_fake_quant(x, rule, op_id)
      out = norm(x, *args, **kwargs)
    return self._fake_quant_output(out, rule)


class Softmax(QuantizedOp):
  """Softmax op."""

  input_idx = [0]
  # The converter requires (scale, zero_point) = (1.0 / 256.0, -128). Qwix uses
  # [-128, 127], which maps to [0, 255 / 256] with the above scale/zp.
  fixed_range_for_output = (0.0, 255 / 256)


class Tanh(QuantizedOp):
  """tanh op."""

  input_idx = [0]
  # The converter requires (scale, zero_point) = (1.0 / 128.0, 0). Qwix uses
  # [-128, 127], which maps to [-1, 127 / 128] with the above scale/zp.
  fixed_range_for_output = (-1.0, 127 / 128)


class UfuncCall(QuantizedOp):
  """ufunc call op."""

  def __call__(self, *args: Any, **kwargs: Any) -> jax.Array:
    # Assume single threaded.
    self._op_name = args[0].__name__
    self._output_allow_fusion = False
    self.input_idx = [1] if self._op_name == 'negative' else [1, 2]
    return super().__call__(*args, **kwargs)

  def _fake_quant_inputs(
      self,
      args: Sequence[Any],
      rule: qconfig.QuantizationRule | None,
      op_id: str,
  ) -> Sequence[Any]:
    """Fake quantize the inputs of the op."""
    if (
        self._op_name in ('add', 'sub', 'mul', 'truediv')
        and aux_data.get(args[1], _ALLOW_FUSION, False)
        and not aux_data.get(args[2], _IS_ACTIVATION, False)
    ):
      # The previous op allows to fuse adding a constant.
      self._output_allow_fusion = True
      return args
    return super()._fake_quant_inputs(args, rule, op_id)

  def _fake_quant_output(
      self, array: jax.Array, rule: qconfig.QuantizationRule | None
  ) -> jax.Array:
    if self._output_allow_fusion:
      aux_data.set(array, _ALLOW_FUSION, True)
    return super()._fake_quant_output(array, rule)


class Concatenate(QuantizedOp):
  """Concatenate op."""

  def __call__(self, arrays: Sequence[jax.Array], *args, **kwargs) -> jax.Array:
    """QAT concatenate."""
    if not any(aux_data.get(x, _IS_ACTIVATION, False) for x in arrays):
      return self._call_original_op(arrays, *args, **kwargs)

    # Forward the fixed_range if all inputs have the same.
    fixed_range = aux_data.get(arrays[0], _FIXED_RANGE, None)
    if any(aux_data.get(x, _FIXED_RANGE, None) != fixed_range for x in arrays):
      fixed_range = None

    # If ourselves is not quantized, fake quantize the inputs if needed.
    # Otherwise, don't insert FQ for the inputs.
    rule, op_id = self._get_rule_and_op_id_fn(self._op_name)
    if not rule or rule.act_qtype is None:
      arrays = [
          self._maybe_fake_quant(x, rule, op_id + f'_input{i}')
          for i, x in enumerate(arrays)
      ]

    out = jnp.concatenate(arrays, *args, **kwargs)

    if fixed_range is not None:
      aux_data.set(out, _FIXED_RANGE, fixed_range)
    return self._fake_quant_output(out, rule)


class Take(OnlyInputOp):
  """jax.numpy.take or rewriting_take."""

  input_idx = [0, 1]

  def __call__(self, *args, **kwargs) -> jax.Array:
    if not self._inputs_have_activations(args):
      return self._call_original_op(*args, **kwargs)

    # Only x needs to be fake quantized.
    args = list(args)
    rule, op_id = self._get_rule_and_op_id_fn(self._op_name)
    args[0] = self._maybe_fake_quant(args[0], rule, op_id)

    if rule and rule.act_qtype:
      # Provide a default fill_value for take. Otherwise, it will be nan
      # and cause the conversion to fail.
      kwargs.setdefault('fill_value', 0)

    out = self._call_original_op(*args, **kwargs)
    if rule and rule.act_qtype:
      # Output doesn't need more FQ.
      aux_data.set(out, _FQ_ARRAY, 'self')
    return self._fake_quant_output(out, rule)


class Split(OnlyInputOp):
  """jax.lax.split."""

  def __call__(
      self, x: jax.Array, sizes: Sequence[int], axis: int = 0
  ) -> jax.Array:
    if not aux_data.get(x, _IS_ACTIVATION, False):
      return self._call_original_op(x, sizes, axis)

    rule, op_id = self._get_rule_and_op_id_fn(self._op_name)
    x = self._maybe_fake_quant(x, rule, op_id)

    outputs = self._call_original_op(x, sizes, axis)
    # Output doesn't need more FQ.
    for output in outputs:
      aux_data.set(output, _FQ_ARRAY, 'self')
      self._fake_quant_output(output, rule)
    return outputs


class Silu(QuantizedOp):
  """jax.nn.silu."""

  def __call__(self, x: jax.Array) -> jax.Array:
    """QAT silu."""
    if not aux_data.get(x, _IS_ACTIVATION, False):
      return self._call_original_op(x)
    rule, op_id = self._get_rule_and_op_id_fn(self._op_name)
    x = self._maybe_fake_quant(x, rule, op_id)
    y = jax.nn.sigmoid(x)
    aux_data.set(y, _FIXED_RANGE, Softmax.fixed_range_for_output)
    y = self._maybe_fake_quant(y, rule, op_id + '_sigmoid')
    return self._fake_quant_output(x * y, rule)


class DotEinsumConv(QuantizedOp):
  """dot_general/einsum/conv_general_dilated."""

  # Whether to check if the lhs is an activation.
  check_activation: bool = True

  # Whether to disable per-channel quantization for weights.
  disable_per_channel_weights: bool = False

  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    if self._op_name == 'einsum':
      self.input_idx = [1, 2]
    else:
      self.input_idx = [0, 1]

  def _get_how_to_quantize(
      self,
      for_lhs: bool,
      qtype: jax.typing.DTypeLike,
      calibration_method: str,
      args: Sequence[Any],
      kwargs: dict[str, Any],
  ) -> qarray.HowToQuantize:
    """Get the HowToQuantize for the given op and arguments."""
    match self._op_name:
      case 'dot_general' | 'dot':
        if self._op_name == 'dot':
          assert args[1].ndim <= 2
          d_num = ((args[0].ndim - 1,), (0,)), ((), ())
        else:
          d_num = args[2] if len(args) > 2 else kwargs['dimension_numbers']
        return dot_general.get_how_to_quantize(
            dimension_numbers=d_num,
            ndims=(len(args[0].shape), len(args[1].shape)),
            for_lhs=for_lhs,
            qtype=qtype,
            tile_size=None,  # Tiling is not supported in ODML.
            calibration_method=calibration_method,
        )
      case 'einsum':
        return einsum.get_how_to_quantize(
            einsum_str=args[0],
            ndims=(len(args[1].shape), len(args[2].shape)),
            for_lhs=for_lhs,
            qtype=qtype,
            tile_size=None,  # Tiling is not supported in ODML.
            calibration_method=calibration_method,
        )
      case 'conv_general_dilated':
        d_num = args[6] if len(args) > 6 else kwargs['dimension_numbers']
        return conv_general.get_how_to_quantize(
            dimension_numbers=d_num,
            for_lhs=for_lhs,
            qtype=qtype,
            calibration_method=calibration_method,
        )
      case _:
        raise ValueError(f'Unsupported op {self._op_name}.')

  def __call__(self, *args, **kwargs) -> jax.Array:
    lhs_idx, rhs_idx = self.input_idx
    rule, op_id = self._get_rule_and_op_id_fn(self._op_name)
    args = list(args)

    lhs_is_activation = aux_data.get(args[lhs_idx], _IS_ACTIVATION, False)
    lhs_is_weight = aux_data.get(args[lhs_idx], _WEIGHT_NAME, None) is not None
    rhs_is_activation = aux_data.get(args[rhs_idx], _IS_ACTIVATION, False)
    rhs_is_weight = aux_data.get(args[rhs_idx], _WEIGHT_NAME, None) is not None
    assert lhs_is_activation + lhs_is_weight <= 1
    assert rhs_is_activation + rhs_is_weight <= 1

    # Check for strict mode.
    if self.check_activation and not lhs_is_activation and not lhs_is_weight:
      raise NotAnActivationError

    # Fake quantize lhs.
    if (rule and rule.act_qtype and not rule.act_static_scale) and (
        lhs_is_activation and rhs_is_weight  # DRQ only supports act x weight.
    ):
      # Handle DRQ, which allows per-channel quantization for activations.
      lhs_how = self._get_how_to_quantize(
          for_lhs=True,
          qtype=rule.act_qtype,
          calibration_method=rule.act_calibration_method,
          args=args,
          kwargs=kwargs,
      )
      args[lhs_idx] = self._fake_quant_fn(args[lhs_idx], lhs_how, None)
    else:
      # Assume lhs is an activation or a weight but not a constant.
      args[lhs_idx] = self._maybe_fake_quant(
          args[lhs_idx], rule, op_id + '_lhs'
      )

    # Fake quantize rhs.
    if rule and rule.weight_qtype and rhs_is_weight:
      # Weights on RHS can be per-channel quantized.
      rhs_how = self._get_how_to_quantize(
          for_lhs=False,
          qtype=rule.weight_qtype,
          calibration_method=rule.weight_calibration_method,
          args=args,
          kwargs=kwargs,
      )
      if self.disable_per_channel_weights:
        rhs_how = dataclasses.replace(rhs_how, channelwise_axes=())
      args[rhs_idx] = self._fake_quant_fn(args[rhs_idx], rhs_how, None)
    elif rhs_is_activation:
      args[rhs_idx] = self._maybe_fake_quant(
          args[rhs_idx], rule, op_id + '_rhs'
      )

    out = self._call_original_op(*args, **kwargs)
    aux_data.set(out, _ALLOW_FUSION, True)
    return self._fake_quant_output(out, rule)


class CustomJvpCall(OnlyOutputOp):
  """This is only for intercepting relu and relu6.

  It's possible to add generic jax.custom_jvp support to interception.py, but we
  only have relu that needs to be intercepted in this way, so let's keep it
  simple for now.
  """

  input_idx = [1]

  def __call__(self, *args: Any, **kwargs: Any) -> Any:
    if args[0] in (jax.nn.relu, jax.nn.relu6):
      return super().__call__(*args, **kwargs)
    return self._call_original_op(*args, **kwargs)



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Support feeding QArray into Pallas kernels."""

import copy
import dataclasses
import functools
from typing import Any, Callable

import jax
from jax.experimental import pallas as pl
import numpy as np
from qwix._src.core import qarray


def pallas_call(
    kernel: Callable[..., None],
    out_shape: Any,
    *,
    grid_spec: pl.GridSpec | None = None,
    grid=(),
    in_specs=pl.no_block_spec,
    out_specs=pl.no_block_spec,
    scratch_shapes=(),
    **kwargs,
) -> Callable[..., Any]:
  """A lifted version of jax.pallas_call that takes QArray as arguments.

  The specs for the QArray arguments should be the same as the original jax
  arrays, and this function will take care of converting them properly.

  Args:
    kernel: same as in jax.pallas_call.
    out_shape: same as in jax.pallas_call.
    grid_spec: same as in jax.pallas_call.
    grid: same as in jax.pallas_call.
    in_specs: same as in jax.pallas_call. The specs for QArrays should stay the
      same as the original jax arrays.
    out_specs: same as in jax.pallas_call.
    scratch_shapes: same as in jax.pallas_call.
    **kwargs: extra arguments to pass to jax.pallas_call.

  Returns:
    A function that can be called on a number of positional array arguments to
    invoke the Pallas kernel.
  """
  if grid_spec is None:
    grid_spec = pl.GridSpec(grid, in_specs, out_specs, scratch_shapes)

  def wrapper(*args):
    num_scalar_prefetch = getattr(grid_spec, "num_scalar_prefetch", 0)

    in_specs = update_block_specs_for_qarray(
        grid_spec.in_specs, args[num_scalar_prefetch:]
    )
    in_specs, new_args, restore_fn = transform_block_specs_for_tpu(
        in_specs, args[num_scalar_prefetch:]
    )
    args = args[:num_scalar_prefetch] + new_args

    # pl.GridSpec doesn't support dataclasses.replace.
    new_grid_spec = copy.copy(grid_spec)
    new_grid_spec.in_specs = in_specs

    return pl.pallas_call(
        lambda *kernel_args: kernel(
            *kernel_args[:num_scalar_prefetch],
            *restore_fn(kernel_args[num_scalar_prefetch : len(args)]),
            *kernel_args[len(args) :],
        ),
        out_shape,
        grid_spec=new_grid_spec,
        **kwargs,
    )(*args)

  return wrapper


def update_block_specs_for_qarray(block_specs: Any, args: Any) -> Any:
  """Update block specs for QArray arguments."""

  def _update_block_spec(spec: pl.BlockSpec, arg):
    if not isinstance(arg, qarray.QArray):
      return spec

    # Calculate block size of the scale array for each axis.
    scale_block_shape = []
    for v, bv, s in zip(arg.qvalue.shape, spec.block_shape, arg.scale.shape):
      if bv is None:
        scale_block_shape.append(None)
      elif s == 1:
        scale_block_shape.append(1)
      else:
        assert v % bv == 0 and s % (v // bv) == 0, f"{v=} {bv=} {s=}"
        scale_block_shape.append(s // (v // bv))
    scale_block_shape = tuple(scale_block_shape)

    scale_index_map = lambda *a: tuple(
        i if s > 1 else 0 for i, s in zip(spec.index_map(*a), arg.scale.shape)
    )
    scale_block_spec = dataclasses.replace(
        spec, block_shape=scale_block_shape, index_map=scale_index_map
    )
    assert arg.zero_point is None, "Zero point is not supported yet."

    return dataclasses.replace(arg, qvalue=spec, scale=scale_block_spec)

  return jax.tree.map(_update_block_spec, block_specs, args)


def transform_block_specs_for_tpu(
    block_specs: Any, args: Any
) -> tuple[Any, Any, Callable[..., Any]]:
  """Transform block specs and arguments so that they can be used on TPU.

  This workarounds the Pallas TPU requirement that the block shapes must be
  divisible by 8x128. A "restore" function is returned to restore the actual
  unpadded block shapes inside the kernel.

  Args:
    block_specs: a pytree of pl.BlockSpec.
    args: the arguments to the pallas_call, which have the same structure as
      block_specs.

  Returns:
    A tuple of (new_block_specs, new_args, restore_fn), where new_block_specs
    and new_args are the transformed pytrees, and restore_fn is a function that
    needs to be called inside the kernel on the kernel arguments to restore the
    actual block shapes.
  """
  flatten_block_specs, treedef = jax.tree.flatten(block_specs)
  flatten_args = treedef.flatten_up_to(args)

  # Information needed to restore the original block shapes inside the kernel.
  # The keys are the indices of the arrays in the flattened pytree. A key may
  # only appear in one of the dictionaries.
  reverse_transposes = {}
  reverse_reshapes = {}

  for i, (spec, arg) in enumerate(zip(flatten_block_specs, flatten_args)):
    # Check if the block shape is already optimal.
    if _is_optimal_for_tpu(spec.block_shape, arg.shape):
      continue

    # Solution 1: try to transpose the array to put the longest axis at the end.
    transpose = np.argsort([1 if s is None else s for s in spec.block_shape])
    block_shape_t = _reorder(spec.block_shape, transpose)
    if _can_fit_tpu_requirements(block_shape_t, _reorder(arg.shape, transpose)):
      flatten_args[i] = arg.transpose(transpose)
      index_map_t = functools.partial(
          lambda spec, transpose, *a: _reorder(spec.index_map(*a), transpose),
          spec,
          transpose,
      )
      flatten_block_specs[i] = dataclasses.replace(
          spec, block_shape=block_shape_t, index_map=index_map_t
      )
      reverse_transposes[i] = np.argsort(
          [t for t in transpose if spec.block_shape[t] is not None]
      )
      continue

    # Solution 2: reshape the array into (*num_blocks, 1, prod(block_shape))).
    # This satisfies the TPU requirement because the last two dimensions are
    # equal to the respective dimension of the overall array.
    dims = range(arg.ndim)
    arg_t = qarray.split_axis(arg, {i: spec.block_shape[i] or 1 for i in dims})
    arg_t = arg_t.transpose([2 * i for i in dims] + [2 * i + 1 for i in dims])
    arg_t = arg_t.reshape([arg_t.shape[i] for i in dims] + [1, -1])
    flatten_args[i] = arg_t
    block_shape_t = tuple([1 for _ in dims] + [1, arg_t.shape[-1]])
    index_map_t = functools.partial(
        lambda spec, *a: spec.index_map(*a) + (0, 0), spec
    )
    flatten_block_specs[i] = dataclasses.replace(
        spec, block_shape=block_shape_t, index_map=index_map_t
    )
    reverse_reshapes[i] = [bs for bs in spec.block_shape if bs is not None]

  def restore(kernel_args):
    kernel_args = treedef.flatten_up_to(kernel_args)
    for i, kernel_arg in enumerate(kernel_args):
      if i in reverse_transposes:
        # Use pallas-friendly transpose.
        kernel_args[i] = qarray.transpose_array(
            kernel_arg[...], reverse_transposes[i]
        )
      elif i in reverse_reshapes:
        kernel_args[i] = kernel_arg[...].reshape(reverse_reshapes[i])
    return treedef.unflatten(kernel_args)

  return (
      treedef.unflatten(flatten_block_specs),
      treedef.unflatten(flatten_args),
      restore,
  )


def _reorder(
    sequence: tuple[Any, ...], order: tuple[int, ...]
) -> tuple[Any, ...]:
  """Reorder/transpose a sequence of elements."""
  return tuple(sequence[i] for i in order)


def _can_fit_tpu_requirements(
    block_shape: tuple[int | None, ...], arg_shape: tuple[int, ...]
) -> bool:
  """Check if the block shape can fit the TPU requirements."""
  block_shape = tuple(1 if s is None else s for s in block_shape)
  return (block_shape[-1] % 128 == 0 or block_shape[-1] == arg_shape[-1]) and (
      block_shape[-2] % 8 == 0 or block_shape[-2] == arg_shape[-2]
  )


def _is_optimal_for_tpu(
    block_shape: tuple[int | None, ...], arg_shape: tuple[int, ...]
) -> bool:
  """Check if the block shape is already optimal for TPU."""
  block_shape = tuple(1 if s is None else s for s in block_shape)
  if not _can_fit_tpu_requirements(block_shape, arg_shape):
    # Not optimal if cannot fit the TPU requirements.
    return False
  if block_shape[-1] % 128 == 0 and block_shape[-2] % 8 == 0:
    # Optimal if no padding is needed.
    return True
  # Optimal if the block shape is already sorted.
  return sorted(block_shape) == list(block_shape)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Numerics for quantization."""

from typing import Callable, Sequence
import jax
from jax import numpy as jnp

# A function that generates noise for stochastic rounding.
# args:
#   shape: The requested shape of the noise to generate.
# returns: Noise as a jax.Array whose shape is broadcastable to the requested
#   shape, and whose dtype can be promoted to fp32 implicitly.
NoiseFn = Callable[[Sequence[int]], jax.Array]


def should_quantize(dtype: jax.typing.DTypeLike) -> bool:
  """Returns True if the dtype should be quantized."""
  return jnp.dtype(dtype) in [jnp.bfloat16, jnp.float32]


def can_dequant_on_output(qtype: jax.typing.DTypeLike) -> bool:
  """qtypes that cannot be optimized by computing in quantized types first then dequantize."""
  return qtype not in ['nf4']


def get_asymmetric_bound(qtype: jax.typing.DTypeLike) -> tuple[float, float]:
  """Returns the bound of the given qtype used in asymmetric quantization."""
  try:
    dtype = jnp.dtype(qtype)
  except TypeError as e:
    raise ValueError(f"{qtype} doesn't support asymmetric quantization.") from e

  match dtype:
    case jnp.int8:
      return (-128.0, 127.0)
    case jnp.int4:
      return (-8.0, 7.0)
    case _:
      raise ValueError(f"{qtype} doesn't support asymmetric quantization.")


def get_symmetric_bound(qtype: jax.typing.DTypeLike) -> float:
  """Returns the bound of the given qtype used in symmetric quantization."""
  match qtype:
    case 'nf4':
      return 1.0
    case 'int2' | 'int3' | 'int5' | 'int6' | 'int7':
      # The bound is extended to qmax + 0.5 so that we have a better utilization
      # of the qmax bucket. This is more important for fewer bits of int.
      return 2 ** (int(qtype[3:]) - 1) - 0.5
    case 'mxfp8':
      qtype = jnp.float8_e4m3fn
    case 'mxfp4':
      qtype = jnp.float4_e2m1fn

  # Prevent common misconfigurations, e.g., use bf16 as qtype.
  if jnp.dtype(qtype).itemsize > 1:
    raise ValueError(f'Cannot use {qtype} as qtype.')
  try:
    # TODO(dangyi): Extend the finfo.max bucket for a better utilization.
    return float(jnp.finfo(qtype).max)
  except ValueError:
    # See the comment above for why we add 0.5.
    return jnp.iinfo(qtype).max + 0.5


def convert_to(
    x: jax.Array,
    qtype: jax.typing.DTypeLike,
    noise_fn: NoiseFn | None = None,
) -> jax.Array:
  """Rounds and converts x to the given qtype."""
  # Handles synthetic qtypes.
  match qtype:
    case 'nf4':
      return fp_to_nf4(x)
    case 'int2' | 'int3' | 'int5' | 'int6' | 'int7':
      bits = int(qtype[3:])
      qmin = -(2 ** (bits - 1))
      qmax = 2 ** (bits - 1) - 1
      if bits <= 4:
        qtype = jnp.int4
      elif bits <= 8:
        qtype = jnp.int8
      else:
        raise ValueError(f'Unsupported integer dtype: {qtype}')
      return jnp.round(x).clip(qmin, qmax).astype(qtype)
    case 'mxfp8':
      qtype = jnp.float8_e4m3fn
    case 'mxfp4':
      qtype = jnp.float4_e2m1fn

  # Handles builtin qtypes.
  try:
    finfo = jnp.finfo(qtype)
  except ValueError:
    pass
  else:
    # dtype is a floating point type. No rounding needed, but we need to clip to
    # the range to avoid inf or nan (e.g. for e4m3fn).
    return x.clip(float(finfo.min), float(finfo.max)).astype(qtype)

  # dtype is an integer type. We need to round manually but clipping can be
  # handled by "astype".
  if noise_fn is not None:
    # Stochastic rounding is done in fp32 to avoid bias from bf16, e.g.
    # round(bf16(41)-bf16(0.4)) ~= round(40.5) = 40, rather than
    # round(41-0.4) = round(40.6) = 41.
    x = x.astype(jnp.float32) + noise_fn(x.shape)
  return jnp.round(x).astype(qtype)


def convert_from(x: jax.Array, qtype: jax.typing.DTypeLike) -> jax.Array:
  """Converts x from the given qtype."""
  match qtype:
    case 'nf4':
      return nf4_to_fp(x)
    case _:
      # For native types, no extra conversion is needed. The dtype will be
      # converted during unquantization.
      return x


### NF4


# NB: to work around the issue of calling Jax functions in module-level context.
def get_nf4_buckets() -> jax.Array:
  """Returns the NF4 buckets defined in Appendix E in https://arxiv.org/pdf/2305.14314."""
  nf4_buckets = jnp.array([
      -1.0,
      -0.6961928009986877,
      -0.5250730514526367,
      -0.39491748809814453,
      -0.28444138169288635,
      -0.18477343022823334,
      -0.09105003625154495,
      0.0,
      0.07958029955625534,
      0.16093020141124725,
      0.24611230194568634,
      0.33791524171829224,
      0.44070982933044434,
      0.5626170039176941,
      0.7229568362236023,
      1.0,
  ])
  return nf4_buckets


def fp_to_nf4(array: jax.Array) -> jax.Array:
  """Quantizes an array to a 4-bit NormalFloat representation."""
  nf4_buckets = get_nf4_buckets()

  def bucketize(x):
    bucket = jnp.argmin(jnp.abs(nf4_buckets - x))
    return bucket

  buckets = jax.vmap(bucketize)(array.ravel())
  return buckets.astype(jnp.uint4).reshape(array.shape)  # stored as uint4.


def nf4_to_fp(array: jax.Array) -> jax.Array:
  """Dequantizes a NF4 array to original dtype."""
  nf4_buckets = get_nf4_buckets()

  def reverse_bucketize(x):
    return nf4_buckets[x]

  return jax.vmap(reverse_bucketize)(array.ravel()).reshape(array.shape)



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Model-level quantization config."""

from collections.abc import Callable, Collection, Sequence
import dataclasses
import re
from typing import Any

from absl import logging
from flax import nnx
import jax
from jax.experimental import pallas as pl
from qwix._src import aux_data
from qwix._src import flax_util
from qwix._src import interception


@dataclasses.dataclass(frozen=True, kw_only=True)
class QuantizationRule:
  """Quantization rules that match and configure the quantization behavior."""

  ########################################################
  ### "Keys" that are used to match modules.
  ########################################################

  # A regex for matching the module path.
  module_path: str = '.*'

  # The ops this rule applies to, which is a collection of op names, e.g.,
  # einsum. If not set, the rule applies to all ops.
  op_names: Collection[str] = ()

  ########################################################
  ### "Configs" that specify the quantization behavior.
  ########################################################

  # Quantized type for weights.
  weight_qtype: jax.typing.DTypeLike | None = None

  # If set, quantize the activations to the given type.
  act_qtype: jax.typing.DTypeLike | None = None

  # If set, enable subchannel for the contraction axis with the given tile size.
  # If it's a float, it must be "1 / tile_count" and the actual tile size will
  # be round(axis_size * tile_size).
  tile_size: int | float | None = None

  # If set, enable static-range quantization for input activations. Quantization
  # statistics for the input activations are collected and will appear in the
  # Flax collection "quant_stats". This requires act_qtype to be set.
  #
  # If None, the default behavior is to use static-range quantization in ODML
  # and dynamic-range quantization in XLA.
  act_static_scale: bool | None = None

  # The method to calibrate weights, in format of <method>[,<args>]. Supported
  # methods are:
  #   absmax[,<scale>]: symmetric quantization using maximum absolute value. A
  #     scale factor is optional, e.g. "absmax,0.8" clips the range to 0.8x.
  #   minmax[,<scale>]: asymmetric quantization using min and max values. A
  #     scale factor is optional.
  #   rms,<scale>: symmetric quantization using root mean square. A scale factor
  #     is required for this method.
  #   fixed,<min>,<max>: use a fixed range [min, max], e.g. "fixed,-10,10". The
  #     quantization is symmetric if min == -max.
  weight_calibration_method: str = 'absmax'

  # The method to calibrate activations. Supported methods are the same as
  # weight_calibration_method. If None, the default behavior is to use minmax
  # in ODML and absmax in XLA.
  act_calibration_method: str | None = None

  # Batch axes when calculating scales for activations. This only applies when
  # static-range quantization is enabled. Batch axes are treated differently
  # during calibration, i.e., the mean of quant stats is calculated along the
  # batch axes. In dynamic-range quantization, the batch axes will have
  # per-channel scales and this config is ignored.
  act_batch_axes: Collection[int] = (0,)


def get_current_rule(op_name: str) -> QuantizationRule | None:
  """Returns the current quantization rule if intercepted, or None otherwise."""
  del op_name
  return None


class QuantizationProvider:
  """Interface for model integration.

  A provider can be either explicitly called by model authors, or implicitly
  injected into the model by using interception.py.
  """

  def __init__(self, rules: Sequence[QuantizationRule]):
    """Initialize the provider.

    Args:
      rules: The quantization rules in the order of precedence.
    """
    self._rules = [self._init_rule(rule) for rule in rules]
    self._logged_ops = set()

  def _init_rule(self, rule: QuantizationRule) -> QuantizationRule:
    """Validate and set default values for the rule."""
    if rule.act_qtype is None and rule.act_static_scale is not None:
      raise ValueError(f'Invalid rule: {rule}.')
    if rule.act_static_scale is None:
      rule = dataclasses.replace(rule, act_static_scale=False)
    if rule.act_calibration_method is None:
      rule = dataclasses.replace(rule, act_calibration_method='absmax')
    return rule

  def get_intercept_map(self) -> dict[str, Callable[..., Any]]:
    """Returns the intercept map for interception.wrap_func_intercepted."""
    # Common functions that are intercepted by all quantization providers.
    intercept_map = {
        'qwix._src.qconfig.get_current_rule': (
            lambda op: self._get_current_rule_and_op_id(op, only_rule=True)[0]
        )
    }
    if interception.has_attribute('jax.experimental.pallas.pallas_call'):
      # Disable interception for ops in pallas_call.
      intercept_map['jax.experimental.pallas.pallas_call'] = (
          lambda *args, **kwargs: interception.disable_interceptions(
              pl.pallas_call(*args, **kwargs)
          )
      )
    return intercept_map

  def process_model_inputs(
      self, model: Any, model_args: Sequence[Any], model_kwargs: dict[str, Any]
  ) -> tuple[Any, Sequence[Any], dict[str, Any]]:
    """Process the model and its inputs before it is called."""
    if isinstance(model, nnx.Module):
      for _, node in model.iter_modules():
        # Clear the op_count which is used in _get_current_rule_and_op_id below.
        aux_data.clear(node)
    return model, model_args, model_kwargs

  def process_model_output(self, method_name: str, model_output: Any) -> Any:
    """Process the model output before it is returned."""
    del method_name
    return model_output

  def _get_current_rule_and_op_id(
      self,
      op_name: str,
      *,
      only_rule: bool = False,
      repeated_call: bool = False,
  ) -> tuple[QuantizationRule | None, str | None]:
    """Returns the quantization rule and a unique op id for given op.

    This function is intended to be called by subclasses.

    Args:
      op_name: The name of the op.
      only_rule: If True, only return the rule and not the op id.
      repeated_call: If True, use the previous op id. This is useful when this
        function is called multiple times for the same op.

    Returns:
      A tuple of the quantization rule and a unique op id for the given op.
    """
    # Lookup the rule.
    module_path = '/'.join(map(str, flax_util.get_current_module_path()))
    rule_idx = None
    for idx, rule in enumerate(self._rules):
      if re.fullmatch(rule.module_path, module_path) and (
          not rule.op_names or op_name in rule.op_names
      ):
        rule_idx = idx
        break
    rule = self._rules[rule_idx] if rule_idx is not None else None
    if only_rule:
      return rule, None

    # Always generate op_id regardless of whether a rule is found.
    module = flax_util.get_current_module()
    count = aux_data.get(module, op_name + '_count', 0)
    if repeated_call:
      count -= 1
    else:
      aux_data.set(module, op_name + '_count', count + 1)
    op_id = op_name + str(count)

    if (module_path, op_id) not in self._logged_ops:
      # Avoid logging the same message multiple times.
      self._logged_ops.add((module_path, op_id))
      logging.info(
          '[QWIX] module=%r op=%s rule=%s', module_path, op_id, rule_idx
      )
    return rule, op_id



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""GPTQ algorithm.

This is a JAX implementation of the GPTQ algorithm from:
https://arxiv.org/pdf/2210.17323

The original implementation is in PyTorch and can be found here:
https://github.com/IST-DASLab/gptq/blob/main/gptq.py
"""

# We try to use the same naming as in the PyTorch implementation, thus
# pylint: disable=invalid-name

import jax
import jax.numpy as jnp
from qwix._src.core import qarray


def cholesky_inverse(L: jax.Array) -> jax.Array:
  return jax.scipy.linalg.cho_solve((L, True), jnp.eye(L.shape[0]))


def find_params(
    w: jax.Array, how: qarray.HowToQuantize
) -> tuple[jax.Array, jax.Array | None]:
  """Finds the optimal quantization parameters for a given weight tensor.

  Args:
    w: weight matrix of shape (rows, groupsize).
    how: how to quantize this group of weights. tiled_axes should either be
      empty or {1: groupsize}.

  Returns:
    A tuple of scale and optional zero_point, which should have shape (1, 1) or
    (rows, 1) when channelwise quantization is enabled.
  """
  calibration = qarray.calibrate(w, how)
  return qarray.compute_scale_zero_point(calibration, how.qtype)


def quantize(
    w: jax.Array,
    qtype: jax.typing.DTypeLike,
    scale: jax.Array,
    zero_point: jax.Array | None,
) -> tuple[jax.Array, jax.Array]:
  """Quantize w and return the raw quantized and dequantized w.

  Args:
    w: weight matrix of shape (rows, 1).
    qtype: The target quantized data type (e.g., jnp.int8).
    scale: The quantization scale returned by find_params.
    zero_point: The quantization zero point (optional) returned by find_params.

  Returns:
    A tuple of:
      - The raw quantized integer values.
      - The dequantized floating-point values (approximation of original w).
  """
  qw = qarray.quantize_with_scale_zero_point(w, qtype, scale, zero_point)
  return qw.qvalue, qarray.dequantize(qw)


def quantize_weight(
    W: jax.Array,
    H: jax.Array,
    how: qarray.HowToQuantize,
    blocksize: int = 128,
    percdamp: float = 0.01,
) -> tuple[qarray.QArray, jax.Array]:
  """Quantize a weight matrix using GPTQ.

  This function minimizes Loss = ||W @ X - W_q @ X||^2 using the Hessian as
  H = X @ X.T.

  Args:
    W: weight matrix in W @ X with shape (rows, columns), where columns is the
      input/contraction dimension and rows is the output dimension.
    H: Hessian of with shape (columns, columns), usually computed as X @ X.T.
    how: how to quantize W.
    blocksize: the GPTQ algorithm blocksize (should usually not be changed)
    percdamp: the percentage of the average diagonal used for dampening H.

  Returns:
    a tuple of (W_q, Losses), where W_q is the quantized weight matrix as a
    QArray and Losses is the overall quantization losses.
  """
  rows, columns = W.shape
  assert H.shape == (columns, columns)

  groupsize = how.tiled_axes.get(1, rows)

  H_diag = jnp.diag(H)
  dead = H_diag == 0
  H = jnp.where(dead & jnp.eye(columns, dtype=bool), 1.0, H)
  W = jnp.where(dead, 0.0, W)

  # Dampen the Hessian.
  damp = percdamp * jnp.mean(H_diag)
  diag = jnp.arange(columns)
  H = H.at[diag, diag].add(damp)

  # Cholesky Inverse for Hessian.
  H = jnp.linalg.cholesky(H)
  H = cholesky_inverse(H)
  H = jnp.linalg.cholesky(H, upper=True)
  Hinv = H

  # Q: the final low-precision integer weights. Will be populated step by step.
  # TODO(dangyi): support synthetic qtype.
  Q = jax.new_ref(jnp.zeros_like(W, dtype=how.qtype))
  # Losses: the quantization error for each element in W. Will be populated step
  # by step.
  Losses = jax.new_ref(jnp.zeros_like(W))

  # scales, zero_points: quantization parameters for each group.
  scales, zero_points = [], []

  # The GPTQ algorithm processes the weight matrix `W` in blocks of columns.
  # In each block, weights are quantized column by column, and an error
  # compensation is applied to the remaining unquantized columns in the block.
  for i1 in range(0, columns, blocksize):
    i2 = min(i1 + blocksize, columns)
    count = i2 - i1

    # Prepare a block of columns from the entire weight matrix and hessian:
    # W1 and Hinv1.
    W1 = jax.new_ref(W[:, i1:i2])
    Err1 = jax.new_ref(jnp.zeros_like(W1))
    Losses1 = jax.new_ref(jnp.zeros_like(W1))
    Hinv1 = Hinv[i1:i2, i1:i2]

    # Process each column within the current block.
    for i in range(count):
      w = W1[:, i]
      d = Hinv1[i, i]

      # Find the quantization parameters for the current group of weights.
      # This is done once every `groupsize` columns.
      if (i1 + i) % groupsize == 0:
        scale, zero_point = find_params(
            W[:, (i1 + i) : (i1 + i + groupsize)], how
        )
        scales.append(scale.squeeze(1))
        zero_points.append(zero_point and zero_point.squeeze(1))

      # Quantize the current column `w` and get the dequantized value `dq`.
      q, dq = quantize(w, how.qtype, scales[-1], zero_points[-1])
      Q[:, i1 + i] = q
      # Calculate the quantization loss for this column.
      Losses1[:, i] = (w - dq) ** 2 / d**2

      # Update all remaining unquantized columns in the current block to
      # compensate for the quantization error introduced in the current column.
      err1 = (w - dq) / d
      W1[:, i:] -= jnp.outer(err1, Hinv1[i, i:])
      Err1[:, i] = err1

    # Accumulate the losses from the current block to the total Losses.
    Losses[:, i1:i2] = Losses1[...] / 2

    # Update the full weight matrix `W`. The columns in the current block are
    # updated with the error-compensated `W1`. The columns in subsequent blocks
    # are also adjusted based on the errors accumulated in `Err1`.
    W = W.at[:, i1:i2].set(W1[...])
    W = W.at[:, i2:].subtract(Err1[...] @ Hinv[i1:i2, i2:])

  # Stack the scales and zero points along the column axis.
  scale = jnp.stack(scales, axis=1)
  zero_point = None
  if None not in zero_points:
    zero_point = jnp.stack(zero_points, axis=1)
  return qarray.QArray(Q[...], scale, zero_point), Losses[...]


def compute_hessian(X: jax.Array) -> jax.Array:
  """Computes the Hessian of the GPTQ objective function.

  The GPTQ algorithm minimizes the squared error of the layer's output, rather
  than just the weights. This specifically relies on the input data X to
  determine which weights are most critical to maintain accurately.

  Derivation:
    1. Objective: Minimize difference between original output (Y = W @ X)
       and quantized output (Y_q = W_q @ X).
          L(W_q) = || W @ X - W_q @ X ||^2

    2. Define error term ∆W = W - W_q:
          L(W_q) = || ∆W @ X ||^2

    3. Express squared Frobenius norm as a Trace (||A||^2 = Trace(A @ A.T)):
          L(W_q) = Trace((∆W @ X) @ (∆W @ X).T)
               = Trace(∆W @ X @ X.T @ ∆W.T)

    4. Identify the Hessian:
      The loss is a quadratic form with respect to ∆W. The second derivative
      (Hessian) is exactly twice the constant matrix in the middle term.
          Hessian = 2 * (X @ X.T)
      By convention, the constant factor of 2 is ignored during implementation.

  Args:
    X: Input data matrix of shape (in_features, n_samples).

  Returns:
    The Hessian matrix of shape (in_features, in_features).
  """
  return X @ X.T



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Intercepts Python functions by patching."""

import functools
import sys
import threading
import types
from typing import Any, Callable, Mapping, TypeAlias

import jax
from qwix._src import aux_data

Function: TypeAlias = Callable[..., Any]

# A dict of { (thread_id, interceptor_id): enabled }. The interception only
# applies when (thread_id, interceptor_id) is in the dict and enabled is True.
#
# The purposes are:
#   * (thread_id) avoids the interception leaking to other threads.
#   * (interceptor_id) avoids the same interceptor being applied twice in the
#     same thread.
#   * (enabled) allows the interceptor to be disabled for recursive calls.
_intercepted_threads: dict[tuple[int, int], bool] = {}


def wrap_func_intercepted(
    func: Function,
    get_intercept_map: Callable[[], Mapping[str, Function]],
    *,
    input_transform: Callable[[Any, Any], tuple[Any, Any]] = lambda *x: x,
    output_transform: Callable[[Any], Any] = lambda x: x,
    should_intercept: Callable[[], bool] = lambda: True,
) -> Function:
  """Wrap a function in a scope where functions in intercept_map are intercepted.

  We're doing a little bit more than just monkey-patching the attributes of the
  objects, including

  * The interception is both thread-local and non-recursive, which means
    * the interception of one thread won't affect another thread.
    * calling an intercepted function inside another intercepted function will
      not trigger the double interception.
    * the original functions will be restored when the replaced functions are
      called.

  * We try to patch the code object of a function rather than the function
    itself. This allows us to patch all the aliases of a function, e.g.
    patching jax.nn.gelu will also affect flax.linen.gelu.

    These can be patched by replacing their code object:
      * global functions, e.g. jax.lax.sin
      * unbound methods, e.g. jax.numpy.ufunc.__call__. These don't have to be
        patched through code objects, but it's hard to distinguish them from
        global functions. These also includes static methods.
    These cannot:
      * callable objects, e.g. a PjitFunction object like jnp.sin.
      * bound methods, e.g. jax.lax.sin_p.bind. They have __code__ attribute,
        but cannot be set. These also includes class methods.
      * functions with freevars.

  * To support patching PjitFunction object, we disable JIT if any of those
    functions are intercepted, and patch the _fun.__code__ attributes instead.
    Note that replacing _fun attribute is not sufficient because it's not
    actually used.

  Args:
    func: The function to wrap.
    get_intercept_map: A function that returns a mapping from function names to
      functions, e.g. {"jax.lax.dot_general": quantized_dot_general}.
    input_transform: A function to transform the input (args and kwargs) of the
      function.
    output_transform: A function to transform the output of the function.
    should_intercept: A predicate to decide whether the interception should be
      applied at all.

  Returns:
    A wrapped function.
  """

  @functools.wraps(func)
  def wrapper(*args, **kwargs):
    this_thread = threading.get_ident()
    # In Python, the id of an instance will change every time! i.e.
    # id(obj.method) != id(obj.method).
    interceptor_id = hash(get_intercept_map)

    # Skip if already intercepted.
    if (this_thread, interceptor_id) in _intercepted_threads:
      return func(*args, **kwargs)

    if not should_intercept():
      return func(*args, **kwargs)

    intercept_map = dict(get_intercept_map())

    # Collect the original attributes used for restoration. This also ensures
    # that the names in the intercept_map are valid.
    original_map = _install({name: None for name in intercept_map})

    # Rewrite PjitFunction object to _fun, which will be further rewritten to
    # code object below.
    need_to_disable_jit = False
    for name, fn in list(original_map.items()):
      if isinstance(fn, jax._src.lib._jax.PjitFunction):  # pylint: disable=protected-access
        original_map[name + "._fun"] = original_map.pop(name)._fun  # pylint: disable=protected-access
        intercept_map[name + "._fun"] = intercept_map.pop(name)
        need_to_disable_jit = True

    # Check if JIT is already disabled.
    if jax.config.jax_disable_jit:
      need_to_disable_jit = False

    # Rewrite functions to code objects. We're still storing functions instead
    # of code objects in original_map and intercept_map here, because we need
    # them to be callable but code objects are not.
    #
    # For example, when patching jax.lax.sin,
    #   original_map: "jax.lax.sin.__code__" => a copy of the original function,
    #                 used for calling the original function.
    #   intercept_map: "jax.lax.sin.__code__" => the intercept function, which
    #                  will be converted to a code object through _fn_to_code.
    for name, fn in list(original_map.items()):
      if isinstance(fn, types.FunctionType) and not fn.__code__.co_freevars:
        # _copy_fn is needed because the original function will be modified
        # in place.
        original_map[name + ".__code__"] = _copy_fn(original_map.pop(name))
        intercept_map[name + ".__code__"] = intercept_map.pop(name)

    # Implement thread-local and non-recursive interception.
    for name, fn in intercept_map.items():
      intercept_map[name] = _wrap_to_call_original_if_needed(
          interceptor_id, original_map[name], fn
      )

    # Convert the functions to code objects if needed before _install.
    # Also check if we accidentally intercept a code object multiple times.
    seen_code_objects = set()
    for name, fn in intercept_map.items():
      if name.endswith(".__code__"):
        intercept_map[name] = _fn_to_code(fn)
        original_map[name] = original_map[name].__code__
        if original_map[name] in seen_code_objects:
          raise ValueError(f"Intercept same code object: {original_map}")
        seen_code_objects.add(original_map[name])

    # Apply the input transform.
    args, kwargs = input_transform(args, kwargs)

    _intercepted_threads[(this_thread, interceptor_id)] = True
    _install(intercept_map)
    if need_to_disable_jit:
      jax.config.update("jax_disable_jit", True)
    try:
      output = func(*args, **kwargs)
    finally:
      if need_to_disable_jit:
        jax.config.update("jax_disable_jit", False)
      _install(original_map)
      _intercepted_threads.pop((this_thread, interceptor_id))

    # Apply the output transform.
    return output_transform(output)

  return wrapper


def _wrap_to_call_original_if_needed(
    interceptor_id: int, original_fn: Function, intercept_fn: Function
) -> Function:
  """Check and decide whether to call the original or the intercepted function."""
  # It's unclear but we cannot return a functools.partial object here, otherwise
  # the test_intercept_class_method will fail.

  @functools.wraps(original_fn)
  def wrapper(*args, **kwargs):
    # Check if we should call the original function.
    key = (threading.get_ident(), interceptor_id)
    if not _intercepted_threads.get(key, False):
      return original_fn(*args, **kwargs)

    # Temporarily disable the interception for recursive calls.
    _intercepted_threads[key] = False
    try:
      return intercept_fn(*args, **kwargs)
    finally:
      _intercepted_threads[key] = True

  return wrapper


def _install(func_map: Mapping[str, Any]) -> dict[str, Any]:
  """Install the functions and return the original function map.

  If the given func_map contains None values, the installation will be skipped.
  This is useful for collecting the original functions.

  Args:
    func_map: A mapping from function names to functions.

  Returns:
    The original function map.
  """
  orig_func_map = {}
  for name, func in func_map.items():
    name_parts = name.split(".")
    if name_parts[0] in sys.modules:
      obj = sys.modules[name_parts[0]]
    else:
      raise ValueError(f"Cannot find module: {name_parts[0]}")
    for attr in name_parts[1:-1]:
      obj = getattr(obj, attr)
    attr = name_parts[-1]
    orig_func_map[name] = getattr(obj, attr)
    if func is not None:
      setattr(obj, attr, func)
  return orig_func_map


def _fn_to_code(fn: Function) -> types.CodeType:
  """Returns a code object that takes 0 freevars."""
  # To replace the code object of a global function, we need to create a new
  # code object that is not a closure, but still be able to remember the
  # original function. The trick is to associate the original function with the
  # code object itself with the aux_data module. When the code object is
  # executed, the code object itself can be accessed through
  # inspect.currentframe().f_code.

  def wrapper(*args, **kwargs):
    # Because the wrapper object can replace any code in other modules, so it
    # has to import the modules here.
    import inspect  # pylint: disable=g-import-not-at-top,redefined-outer-name,reimported
    from qwix._src import aux_data  # pylint: disable=g-import-not-at-top,redefined-outer-name,reimported

    fn = aux_data.get(inspect.currentframe().f_code, "fn")  # pytype: disable=attribute-error
    return fn(*args, **kwargs)

  code = wrapper.__code__.replace()  # this creates a new code object
  aux_data.set(code, "fn", fn)
  return code


def _copy_fn(fn: types.FunctionType) -> types.FunctionType:
  """Constructs a new function object with the same attributes as the given one."""
  fn_copy = types.FunctionType(fn.__code__, fn.__globals__)
  for field in (
      "__name__",
      "__qualname__",
      "__annotations__",
      "__defaults__",
      "__kwdefaults__",
      "__module__",
      "__doc__",
      "__dict__",
  ):
    if hasattr(fn, field):
      setattr(fn_copy, field, getattr(fn, field))
  return fn_copy


def disable_interceptions(fn):
  """Return the function with interceptions disabled when called."""

  @functools.wraps(fn)
  def wrapper(*args, **kwargs):
    this_thread = threading.get_ident()
    enabled_interceptors = []
    for key, enabled in _intercepted_threads.copy().items():
      if key[0] == this_thread and enabled:
        enabled_interceptors.append(key)
        _intercepted_threads[key] = False
    try:
      return fn(*args, **kwargs)
    finally:
      for key in enabled_interceptors:
        _intercepted_threads[key] = True

  return wrapper


def has_attribute(name: str) -> bool:
  """Returns if the module exists and has the attribute."""
  name_parts = name.split(".")
  if name_parts[0] not in sys.modules:
    return False
  obj = sys.modules[name_parts[0]]
  for attr in name_parts[1:]:
    obj = getattr(obj, attr, None)
    if obj is None:
      return False
  return True



# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Quantized jax.lax.dot_general with subchannel support."""

from collections.abc import Collection, Sequence
import itertools
from typing import Any
import jax
from jax import numpy as jnp
from qwix._src.core import numerics
from qwix._src.core import qarray


def get_how_to_quantize(
    *,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    ndims: tuple[int, int],
    for_lhs: bool,
    tile_size: int | float | None,
    **kwargs: Any,
) -> qarray.HowToQuantize:
  """Get how to quantize from dimension_numbers and remaining_dims.

  By default, use channelwise for all non-contraction axes, and subchannel
  for contraction axes if a tile_size is given.

  Args:
    dimension_numbers: The dimension numbers passed to dot_general.
    ndims: The number of dimensions for lhs and rhs.
    for_lhs: Whether to quantize lhs or rhs.
    tile_size: The tile size for subchannel quantization.
    **kwargs: Additional keyword arguments to HowToQuantize.

  Returns:
    How to quantize.
  """
  if for_lhs:
    ndim = ndims[0]
    contracting_axes = dimension_numbers[0][0]
  else:
    ndim = ndims[1]
    contracting_axes = dimension_numbers[0][1]

  channelwise_axes = sorted(set(range(ndim)) - set(contracting_axes))
  tiled_axes = {}
  if tile_size:
    tiled_axes = {contracting_axes[0]: tile_size}

  return qarray.HowToQuantize(
      channelwise_axes=channelwise_axes,
      tiled_axes=tiled_axes,
      **kwargs,
  )


def _get_scale_transpose(
    dimension_numbers: jax.lax.DotDimensionNumbers,
    ndims: tuple[int, int],
) -> tuple[list[int | None], list[int | None]]:
  """Returns the transpose list for lhs_scale and rhs_scale."""
  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers
  lhs_ra = sorted(set(range(ndims[0])) - set(lhs_ca) - set(lhs_ba))
  rhs_ra = sorted(set(range(ndims[1])) - set(rhs_ca) - set(rhs_ba))
  return (
      list(lhs_ba) + list(lhs_ra) + [None] * len(rhs_ra),  # lhs_scale_transpose
      list(rhs_ba) + [None] * len(lhs_ra) + list(rhs_ra),  # rhs_scale_transpose
  )


def _apply_tiling(
    contracting_axes: Sequence[int],
    batch_axes: Sequence[int],
    tiled_axes: Collection[int],
) -> tuple[tuple[int, ...], tuple[int, ...], tuple[int, ...]]:
  """Apply tiling to dimension numbers.

  Each tiled contracting axis is split into two axes, the first being the new
  batch axis, and the second being the new contracting axis.

  Args:
    contracting_axes: The original contracting axes.
    batch_axes: The original batch axes.
    tiled_axes: The tiled axes. Must be a subset of contracting_axes.

  Returns:
    A tuple of (new_ca, new_ba, sum_axes).
  """
  new_ca = [a + sum(t <= a for t in tiled_axes) for a in contracting_axes]
  new_ba = [a + sum(t < a for t in tiled_axes) for a in batch_axes]
  # We choose to insert the tile_count axes to the end of the batch axes.
  # Alternatively, we could insert them to the beginning or to the middle,
  # as long as lhs and rhs use the same order.
  new_ba += [
      a + sum(t < a for t in tiled_axes)
      for a in contracting_axes
      if a in tiled_axes
  ]
  sum_axes = range(len(batch_axes), len(new_ba))
  return tuple(new_ca), tuple(new_ba), tuple(sum_axes)


def _broadcast_axes(
    array: jax.Array, shape: tuple[int, ...], axes: Collection[int]
) -> jax.Array:
  """Broadcast the given axes in the array to the given shape."""
  target_shape = list(array.shape)
  for a in axes:
    target_shape[a] = shape[a]
  return jnp.broadcast_to(array, target_shape)


def _fast_dot_general(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    **kwargs,
) -> jax.Array:
  """Dot general in optimized path by computing in quantized types first then dequantize."""
  if isinstance(lhs, qarray.QArray):
    lhs_value = lhs.qvalue
    lhs_scale = lhs.scale
    lhs_zero_point = lhs.zero_point
    lhs_tiled_axes = qarray.get_tiled_axes(lhs)
  else:
    lhs_value = lhs
    lhs_scale = None
    lhs_zero_point = None
    lhs_tiled_axes = {}
  if isinstance(rhs, qarray.QArray):
    rhs_value = rhs.qvalue
    rhs_scale = rhs.scale
    rhs_zero_point = rhs.zero_point
    rhs_tiled_axes = qarray.get_tiled_axes(rhs)
  else:
    rhs_value = rhs
    rhs_scale = None
    rhs_zero_point = None
    rhs_tiled_axes = {}

  if lhs_zero_point is not None and rhs_zero_point is not None:
    raise ValueError('Only one operand can be asymmetric.')

  (lhs_ca, rhs_ca), (lhs_ba, rhs_ba) = dimension_numbers

  # Figure out the tiled axes to use for the dot_general. For greater
  # flexibility, we allow a non-tiled axis to be contracted with a tiled axis.
  # However, if both axes are tiled, their tile sizes must be the same.
  lhs_tiled_ca = {}
  rhs_tiled_ca = {}
  for l, r in zip(lhs_ca, rhs_ca):
    lhs_tile_size = lhs_tiled_axes.get(l)
    rhs_tile_size = rhs_tiled_axes.get(r)
    if lhs_tile_size and rhs_tile_size and lhs_tile_size != rhs_tile_size:
      raise ValueError(
          'Contracting axes must be tiled with the same tile size.'
          f' {lhs_tiled_axes=} {rhs_tiled_axes=} {dimension_numbers=}'
      )
    if lhs_tile_size or rhs_tile_size:
      lhs_tiled_ca[l] = lhs_tile_size or rhs_tile_size
      rhs_tiled_ca[r] = lhs_tile_size or rhs_tile_size

  # Split lhs/rhs_value for tiled axes.
  lhs_value = qarray.split_axis(lhs_value, lhs_tiled_ca)
  rhs_value = qarray.split_axis(rhs_value, rhs_tiled_ca)

  # Split lhs/rhs_zero_point for tiled axes.
  if lhs_zero_point is not None:
    lhs_zero_point = qarray.split_axis(
        lhs_zero_point, {a: 1 for a in lhs_tiled_ca}
    )
  if rhs_zero_point is not None:
    rhs_zero_point = qarray.split_axis(
        rhs_zero_point, {a: 1 for a in rhs_tiled_ca}
    )

  # Update dimension_numbers and get sum_axes for tiled axes.
  lhs_ca, lhs_ba, sum_axes = _apply_tiling(lhs_ca, lhs_ba, lhs_tiled_ca)
  rhs_ca, rhs_ba, _ = _apply_tiling(rhs_ca, rhs_ba, rhs_tiled_ca)
  dimension_numbers = (lhs_ca, rhs_ca), (lhs_ba, rhs_ba)

  # Transpose lhs/rhs_scale. This works for tiled axes too.
  lhs_scale_transpose, rhs_scale_transpose = _get_scale_transpose(
      dimension_numbers, (len(lhs_value.shape), len(rhs_value.shape))
  )
  if lhs_scale is not None:
    lhs_scale = qarray.split_axis(lhs_scale, {a: 1 for a in lhs_tiled_ca})
    lhs_scale = qarray.transpose_array(lhs_scale, lhs_scale_transpose)
  if rhs_scale is not None:
    rhs_scale = qarray.split_axis(rhs_scale, {a: 1 for a in rhs_tiled_ca})
    rhs_scale = qarray.transpose_array(rhs_scale, rhs_scale_transpose)

  preferred_element_type, result_type = qarray.get_accumulator_and_result_type(
      lhs, rhs, preferred_element_type=preferred_element_type
  )

  res = jax.lax.dot_general(
      lhs_value,
      rhs_value,
      dimension_numbers=dimension_numbers,
      preferred_element_type=preferred_element_type,
      **kwargs,
  )

  if lhs_zero_point is not None:
    # TODO(zhuyunx): This value can be constant folded in SRQ scenarios.
    res = qarray.call_with_generic_broadcast(
        jnp.subtract,
        res,
        jax.lax.dot_general(
            _broadcast_axes(lhs_zero_point, lhs_value.shape, lhs_ca + lhs_ba),
            rhs_value,
            dimension_numbers=dimension_numbers,
            preferred_element_type=preferred_element_type,
            **kwargs,
        ),
    )

  if rhs_zero_point is not None:
    res = qarray.call_with_generic_broadcast(
        jnp.subtract,
        res,
        jax.lax.dot_general(
            lhs_value,
            _broadcast_axes(rhs_zero_point, rhs_value.shape, rhs_ca + rhs_ba),
            dimension_numbers=dimension_numbers,
            preferred_element_type=preferred_element_type,
            **kwargs,
        ),
    )

  if lhs_scale is not None:
    res = qarray.call_with_generic_broadcast(jnp.multiply, res, lhs_scale)
  if rhs_scale is not None:
    res = qarray.call_with_generic_broadcast(jnp.multiply, res, rhs_scale)
  if sum_axes:
    res = jnp.sum(res, axis=sum_axes)
  return res.astype(result_type)


def _slow_dot_general(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    **kwargs,
) -> jax.Array:
  """Dot general in slow path by dequantizing first then computing in floating-point types."""
  if isinstance(lhs, qarray.QArray):
    lhs = qarray.dequantize(lhs)
  if isinstance(rhs, qarray.QArray):
    rhs = qarray.dequantize(rhs)
  return jax.lax.dot_general(lhs, rhs, dimension_numbers, **kwargs)


def loop_dot_general(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    **kwargs,
) -> jax.Array:
  """Loop-based tiled dot general implementation, mainly used in Pallas kernels."""
  if isinstance(lhs, qarray.QArray):
    lhs_value = lhs.qvalue
    lhs_scale = lhs.scale
    assert lhs.zero_point is None
    lhs_tiled_axes = qarray.get_tiled_axes(lhs)
  else:
    lhs_value = lhs
    lhs_scale = None
    lhs_tiled_axes = {}
  if isinstance(rhs, qarray.QArray):
    rhs_value = rhs.qvalue
    rhs_scale = rhs.scale
    assert rhs.zero_point is None
    rhs_tiled_axes = qarray.get_tiled_axes(rhs)
  else:
    rhs_value = rhs
    rhs_scale = None
    rhs_tiled_axes = {}

  lhs_ca, rhs_ca = dimension_numbers[0]

  # Allow non-tiled axes to be contracted with tiled axes.
  ca_tile_counts = []  # number of tiles for each contracting axis.
  for l, r in zip(lhs_ca, rhs_ca):
    lhs_tile_size = lhs_tiled_axes.get(l)
    rhs_tile_size = rhs_tiled_axes.get(r)
    if lhs_tile_size and rhs_tile_size and lhs_tile_size != rhs_tile_size:
      raise ValueError(
          'Contracting axes must be tiled with the same tile size.'
          f' {lhs_tiled_axes=} {rhs_tiled_axes=} {dimension_numbers=}'
      )
    tile_size = lhs_tile_size or rhs_tile_size
    if tile_size is not None:
      ca_tile_counts.append(lhs_value.shape[l] // tile_size)
    else:
      ca_tile_counts.append(1)

  preferred_element_type, result_type = qarray.get_accumulator_and_result_type(
      lhs, rhs, preferred_element_type=preferred_element_type
  )

  lhs_scale_transpose, rhs_scale_transpose = _get_scale_transpose(
      dimension_numbers, (len(lhs_value.shape), len(rhs_value.shape))
  )

  def take_slice(
      array: jax.Array, ca: Sequence[int], ca_tile_indices: Sequence[int]
  ) -> jax.Array:
    indices = []
    for i, s in enumerate(array.shape):
      if i not in ca or s == 1:
        indices.append(slice(None))
        continue
      index = ca_tile_indices[ca.index(i)]
      count = ca_tile_counts[ca.index(i)]
      assert s % count == 0
      size = s // count
      indices.append(slice(index * size, (index + 1) * size))
    return array[tuple(indices)]

  acc = None
  for ca_tile_indices in itertools.product(*map(range, ca_tile_counts)):
    # ca_tile_indices is a list of tile indices for each contracting axis.
    out = jax.lax.dot_general(
        take_slice(lhs_value, lhs_ca, ca_tile_indices),
        take_slice(rhs_value, rhs_ca, ca_tile_indices),
        dimension_numbers=dimension_numbers,
        preferred_element_type=preferred_element_type,
        **kwargs,
    )
    if lhs_scale is not None:
      scale = take_slice(lhs_scale, lhs_ca, ca_tile_indices)
      scale = qarray.transpose_array(scale, lhs_scale_transpose)
      out = qarray.call_with_generic_broadcast(jnp.multiply, out, scale)
    if rhs_scale is not None:
      scale = take_slice(rhs_scale, rhs_ca, ca_tile_indices)
      scale = qarray.transpose_array(scale, rhs_scale_transpose)
      out = qarray.call_with_generic_broadcast(jnp.multiply, out, scale)
    acc = out if acc is None else acc + out
  assert acc is not None
  return acc.astype(result_type)


# If a contracting dimension has a tile size smaller than this threshold, tiled
# dot general will be inefficient and we should dequantize the input first.
MIN_TILE_SIZE_TO_DEQUANT_ON_OUTPUT = 128


def dot_general(
    lhs: qarray.MaybeQArray,
    rhs: qarray.MaybeQArray,
    dimension_numbers: jax.lax.DotDimensionNumbers,
    precision: jax.lax.PrecisionLike = None,
    preferred_element_type: jax.typing.DTypeLike | None = None,
    **kwargs,
) -> jax.Array:
  """Quantized jax.lax.dot_general.

  Args:
    lhs: The left-hand side, either a jax.Array or QArray.
    rhs: The right-hand side, either a jax.Array or QArray.
    dimension_numbers: The dimension numbers passed to dot_general.
    precision: The precision for jax.lax.dot_general.
    preferred_element_type: The preferred element type for jax.lax.dot_general.
    **kwargs: Additional keyword arguments to dot_general.

  Returns:
    a floating-point jax.Array.
  """
  # We need to choose between slow_dot_general, which dequantizes first and
  # then computes in floating-point types, and fast_dot_general, which
  # computes in quantized types first and then dequantize.
  use_fast_dot_general = True
  for operand, ca in zip((lhs, rhs), dimension_numbers[0]):
    if not isinstance(operand, qarray.QArray):
      if numerics.should_quantize(operand.dtype):
        # Always dequantize on inputs if any of the operands is in bf16/fp32,
        # because XLA is able to fuse the dequantize and the matmul. The slow
        # path is usually not slower than the fast path, since both use fp
        # matmul, and will be significantly faster when subchannel or zero_point
        # is used.
        use_fast_dot_general = False
        break
      # For raw arrays in lower precision, e.g. fp8, int4, bool, using fast path
      # may be beneficial.
      continue

    qarray.validate_qarray(operand)

    # qtypes like nf4 cannot be dequantized on output.
    if not numerics.can_dequant_on_output(operand.qtype):
      use_fast_dot_general = False
      break

    # If a contracting dimension is tiled too small, tiled dot general will
    # be inefficient and we should dequantize the input first. This is critical
    # when a contracting dimension is channelwise quantized, e.g. tile_size=1.
    for axis in ca:
      if operand.scale.shape[axis] > 1:
        tile_size = operand.qvalue.shape[axis] // operand.scale.shape[axis]
        if tile_size < MIN_TILE_SIZE_TO_DEQUANT_ON_OUTPUT:
          use_fast_dot_general = False
          break

  if use_fast_dot_general:
    return _fast_dot_general(
        lhs,
        rhs,
        dimension_numbers,
        precision=precision,
        preferred_element_type=preferred_element_type,
        **kwargs,
    )
  else:
    return _slow_dot_general(
        lhs,
        rhs,
        dimension_numbers,
        precision=precision,
        preferred_element_type=preferred_element_type,
        **kwargs,
    )



# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Low-Rank Adapation (LoRA) support."""
import dataclasses
import string
from typing import Callable, Sequence

from flax import linen as nn
from flax import nnx
from flax import typing
import jax
from jax.nn import initializers
from qwix._src import flax_util
from qwix._src import model as qwix_model
from qwix._src import qconfig
from qwix._src.core import einsum
from qwix._src.core import qarray
from qwix._src.providers import ptq


# apply_lora_to_model is just an alias for quantize_model.
apply_lora_to_model = qwix_model.quantize_model


@dataclasses.dataclass(frozen=True, kw_only=True)
class LoraRule(qconfig.QuantizationRule):
  """LoRA rules that match and configure the LoRA behavior."""

  ########################################################
  ### "Configs" that specify the LoRA behavior.
  ########################################################

  # The rank of the LoRA.
  rank: int

  # The alpha scaling parameter of the LoRA. It controls how we update the
  # original weights with LoRA weights
  alpha: float

  # The dropout rate for the LoRA.
  dropout: float = 0.0

  # The initializers for the LoRA A (fan-in) weight, default as he_uniform().
  lora_a_initializer: Callable[..., jax.Array] = initializers.he_uniform()

  # The initializer for the LoRA B (fan-out) weight, default as zeros.
  lora_b_initializer: Callable[..., jax.Array] = initializers.zeros


def _find_lora_dim_char(all_dims: set[str]):
  if 'r' not in all_dims:
    return 'r'
  return sorted(set(string.ascii_letters) - all_dims)[0]


def _parse_einsum_str_for_lora(
    lhs_shape: typing.Shape,
    rhs_shape: typing.Shape,
    einsum_str: str,
    lora_rank: int,
) -> tuple[
    typing.Shape,  # a_shape
    typing.Shape,  # b_shape
    str,  # lora_einsum_str
    Sequence[int | None],  # a_sharding_transpose
    Sequence[int | None],  # b_sharding_transpose
]:
  """Returns lora param shapes and einsum string for LoRA."""
  einsum_info = einsum.get_einsum_info(
      einsum_str, (len(lhs_shape), len(rhs_shape))
  )
  lora_dim_char = _find_lora_dim_char(
      set(einsum_info.lhs) | set(einsum_info.rhs)
  )

  a_shape, b_shape = (), (lora_rank,)
  a_str, b_str = '', lora_dim_char
  a_sharding_transpose, b_sharding_transpose = (), (None,)
  assert len(einsum_info.rhs) == len(rhs_shape)
  for i, (c, dim) in enumerate(zip(einsum_info.rhs, rhs_shape)):
    if c in set(einsum_info.lhs) & set(einsum_info.rhs):  # batch or contracting
      a_str += c
      a_shape += (dim,)
      a_sharding_transpose += (i,)
    else:
      b_str += c
      b_shape += (dim,)
      b_sharding_transpose += (i,)
  a_str += lora_dim_char
  a_shape += (lora_rank,)
  a_sharding_transpose += (None,)

  return (
      a_shape,
      b_shape,
      ','.join([einsum_info.lhs, a_str, b_str]) + '->' + einsum_info.out,
      a_sharding_transpose,
      b_sharding_transpose,
  )


class LoraProvider(ptq.PtqProvider):
  """Provider for (Q)LoRA.

  LoraProvider inherits from PtqProvider, because the base model is frozen
  during LoRA training.
  """

  def __init__(self, rules=None, **kwargs):
    """Initializes the LoraProvider.

    Usage:
      LoraProvider(module_path='module_path', rank=4, alpha=0.5)
    or
      LoraProvider([
          LoraRule(module_path='module_path', rank=4, alpha=0.5)
      ])

    Args:
      rules: A list of quantization rules.
      **kwargs: The keyword arguments to create a rule. Only one of rules and
        kwargs should be provided.
    """
    if rules is None:
      rules = [LoraRule(**kwargs)]
    elif kwargs:
      raise ValueError('Only one of rules and kwargs should be provided.')
    super().__init__(rules=rules)

  def dot_general(
      self,
      lhs: jax.Array,
      rhs: jax.Array | ptq.WithAux[qarray.QArray],
      dimension_numbers: jax.lax.DotDimensionNumbers,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
      out_sharding: jax.sharding.NamedSharding | None = None,
  ) -> jax.Array:
    """LoRA dot_general."""
    res = super().dot_general(
        lhs,
        rhs,
        dimension_numbers,
        precision,
        preferred_element_type,
        out_sharding=out_sharding,
    )

    rule, _ = self._get_current_rule_and_op_id(
        'dot_general', repeated_call=True
    )
    if not isinstance(rule, LoraRule):
      return res

    weight_name = flax_util.find_param(rhs)
    if weight_name is None:  # rhs is not a weight.
      return res

    # We only support ...a,ab->...b for now.
    assert (
        len(rhs.shape) == 2
        and tuple(dimension_numbers[0][1]) == (0,)
        and not dimension_numbers[1][1]
    ), f'Unsupported: {rhs.shape=} {dimension_numbers=}'

    lora_a, lora_b, dropout_layer = _get_or_create_lora_params(
        name=weight_name,
        rule=rule,
        a_shape=(rhs.shape[0], rule.rank),
        b_shape=(rule.rank, rhs.shape[1]),
        a_sharding_transpose=(0, None),
        b_sharding_transpose=(None, 1),
    )

    if dropout_layer is not None:
      # TODO(zhuyunx): Use nnx.Rngs(0) for now. Need to check `deterministic`
      # to decide whether to provide a rng.
      # NOTE: this is wrong because it always uses the same rng.
      lhs = dropout_layer(lhs, rngs=nnx.Rngs(0))

    return res + lhs @ lora_a @ lora_b * (rule.alpha / rule.rank)

  def einsum(
      self,
      einsum_str: str,
      *operands: jax.Array | ptq.WithAux[qarray.QArray],
      **kwargs,
  ) -> jax.Array:
    """LoRA einsum."""
    res = super().einsum(einsum_str, *operands, **kwargs)

    rule, _ = self._get_current_rule_and_op_id('einsum', repeated_call=True)
    if not isinstance(rule, LoraRule):
      return res

    if not isinstance(einsum_str, str) or len(operands) != 2:
      raise ValueError(f'Unsupported einsum format: {einsum_str=} {operands=}')
    lhs, rhs = operands

    weight_name = flax_util.find_param(rhs)
    if weight_name is None:  # rhs is not a weight.
      return res

    (
        a_shape,
        b_shape,
        lora_einsum_str,
        a_sharding_transpose,
        b_sharding_transpose,
    ) = _parse_einsum_str_for_lora(lhs.shape, rhs.shape, einsum_str, rule.rank)

    # Store the lora_einsum_str for debugging.
    module = flax_util.get_current_module()
    setattr(module, weight_name + '_lora_einsum_str', lora_einsum_str)

    lora_a, lora_b, dropout_layer = _get_or_create_lora_params(
        name=weight_name,
        rule=rule,
        a_shape=a_shape,
        b_shape=b_shape,
        a_sharding_transpose=a_sharding_transpose,
        b_sharding_transpose=b_sharding_transpose,
    )

    if dropout_layer is not None:
      # TODO(zhuyunx): Use nnx.Rngs(0) for now. Need to check `deterministic`
      # to decide whether to provide a rng.
      # NOTE: this is wrong because it always uses the same rng.
      lhs = dropout_layer(lhs, rngs=nnx.Rngs(0))

    return res + (
        jax.numpy.einsum(lora_einsum_str, lhs, lora_a, lora_b, **kwargs)
        * (rule.alpha / rule.rank)
    )

  def conv_general_dilated(
      self,
      lhs: jax.Array,
      rhs: jax.Array | ptq.WithAux[qarray.QArray],
      window_strides: Sequence[int],
      padding: str | Sequence[tuple[int, int]],
      lhs_dilation: Sequence[int] | None = None,
      rhs_dilation: Sequence[int] | None = None,
      dimension_numbers: jax.lax.ConvGeneralDilatedDimensionNumbers = None,
      feature_group_count: int = 1,
      batch_group_count: int = 1,
      precision: jax.lax.PrecisionLike = None,
      preferred_element_type: jax.typing.DTypeLike | None = None,
  ) -> jax.Array:
    """LoRA conv_general_dilated."""
    res = super().conv_general_dilated(
        lhs,
        rhs,
        window_strides,
        padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
        precision=precision,
        preferred_element_type=preferred_element_type,
    )

    rule, _ = self._get_current_rule_and_op_id(
        'conv_general_dilated', repeated_call=True
    )
    if not isinstance(rule, LoraRule):
      return res

    weight_name = flax_util.find_param(rhs)
    assert weight_name is not None, 'rhs must be a weight.'

    dimension_numbers = jax.lax.conv_dimension_numbers(
        lhs.shape, rhs.shape, dimension_numbers
    )
    # Assert that the out feature is the last dimension of rhs and out.
    assert (
        dimension_numbers.rhs_spec[0] == len(rhs.shape) - 1
        and dimension_numbers.out_spec[1] == len(lhs.shape) - 1
    ), f'Unsupported: {dimension_numbers=}'

    lora_a, lora_b, dropout_layer = _get_or_create_lora_params(
        name=weight_name,
        rule=rule,
        a_shape=(*rhs.shape[:-1], rule.rank),
        b_shape=(rule.rank, rhs.shape[-1]),
        a_sharding_transpose=(*range(len(rhs.shape) - 1), None),
        b_sharding_transpose=(None, len(rhs.shape) - 1),
    )

    if dropout_layer is not None:
      # TODO(zhuyunx): Use nnx.Rngs(0) for now. Need to check `deterministic`
      # to decide whether to provide a rng.
      # NOTE: this is wrong because it always uses the same rng.
      lhs = dropout_layer(lhs, rngs=nnx.Rngs(0))

    return res + jax.lax.conv_general_dilated(
        lhs,
        lora_a,
        window_strides,
        padding,
        lhs_dilation=lhs_dilation,
        rhs_dilation=rhs_dilation,
        dimension_numbers=dimension_numbers,
        feature_group_count=feature_group_count,
        batch_group_count=batch_group_count,
        precision=precision,
        preferred_element_type=preferred_element_type,
    ) @ lora_b * (rule.alpha / rule.rank)


def _get_or_create_lora_params(
    *,
    name: str,
    rule: LoraRule,
    a_shape: typing.Shape,
    b_shape: typing.Shape,
    a_sharding_transpose: Sequence[int | None],
    b_sharding_transpose: Sequence[int | None],
) -> tuple[jax.Array, jax.Array, nnx.Dropout | None]:
  """Get or create LoRA params.

  Args:
    name: The prefix of the LoRA param.
    rule: The LoRA rule.
    a_shape: The shape of the lora_a param.
    b_shape: The shape of the lora_b param.
    a_sharding_transpose: The transpose to derive the sharding for lora_a.
    b_sharding_transpose: The transpose to derive the sharding for lora_b.

  Returns:
    A tuple of (lora_a, lora_b, dropout_layer).
  """
  dropout_layer = None
  if rule.dropout > 0:
    dropout_layer = nnx.Dropout(rule.dropout)

  # Get the boxed param so that we can access the metadata.
  module = flax_util.get_current_module()
  if isinstance(module, nn.Module):
    param = module.get_variable('params', name)
    lora_a = module.get_variable('params', name + '_lora_a')
    lora_b = module.get_variable('params', name + '_lora_b')
  else:  # isinstance(module, nnx.Module)
    param = getattr(module, name)
    lora_a = getattr(module, name + '_lora_a', None)
    lora_b = getattr(module, name + '_lora_b', None)

  if lora_a is not None and lora_b is not None:
    return flax_util.unbox(lora_a), flax_util.unbox(lora_b), dropout_layer

  def get_canonical_pspec(x):
    """Returns the canonical sharding.spec if x contains a concrete array."""
    x = flax_util.unbox(x)
    sharding = getattr(x, 'sharding', None)
    if not isinstance(sharding, jax.sharding.NamedSharding):
      return None
    # The sharding.spec may be shorter than the ndim.
    padded_pspec = sharding.spec + (None,) * (x.ndim - len(sharding.spec))
    return sharding.update(spec=padded_pspec)

  # Get the dtype, boxed param, and (optional) sharding from the original param.
  if isinstance(param, ptq.WithAux):
    lora_dtype = flax_util.unbox(param.array.scale).dtype
    boxed = param.array.qvalue
    sharding = get_canonical_pspec(boxed)
  else:  # base model is not quantized.
    lora_dtype = flax_util.unbox(param).dtype
    boxed = param
    sharding = get_canonical_pspec(boxed)

  def init(initializer, shape, transpose):
    # TODO(dangyi): use the actual rng.
    value = initializer(jax.random.key(0), shape, lora_dtype)
    if sharding is not None:
      lora_pspec = flax_util.update_sharding(sharding.spec, transpose=transpose)
      value = jax.device_put(value, sharding.update(spec=lora_pspec))
    value = flax_util.update_boxed(boxed, value=value, transpose=transpose)
    if isinstance(value, nnx.Variable):
      return nnx.VariableMetadata(value.value, metadata=value.get_metadata())
    return value

  lora_a = flax_util.get_or_create_param(
      name + '_lora_a',
      lambda: init(rule.lora_a_initializer, a_shape, a_sharding_transpose),
      nnx.LoRAParam,
  )
  lora_b = flax_util.get_or_create_param(
      name + '_lora_b',
      lambda: init(rule.lora_b_initializer, b_shape, b_sharding_transpose),
      nnx.LoRAParam,
  )
  return lora_a, lora_b, dropout_layer
